# -*- coding: utf-8 -*-


from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal

from ogt import ags4, CELL_COLORS

import app_globals as G
from img import Ico
import ags4_widgets
import xwidgets
#import ogtgui_widgets
import ogtgui_dialogs
import ogtgui_errors

from ogt.ogt_validate import OGTError

class HeadersListModel(QtCore.QAbstractTableModel):

    class C:
        #node = 0
        valid = 0
        head_code = 1
        unit = 2
        type = 3
        head_description = 4


    def __init__(self):
        QtCore.QAbstractTableModel.__init__(self)

        self.ogtGroup = None
        self._col_labels = ["Valid", "Heading", "Unit", "Type", "Descripton"]

    def set_group(self, ogtGroup):
        self.ogtGroup = ogtGroup

    def columnCount(self, foo):
        return len(self._col_labels)

    def rowCount(self, midx):

        if self.ogtGroup == None:
            return 0
        return self.ogtGroup.headings_count

    def data(self, midx, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = midx.row()
        col = midx.column()

        if role == Qt.DisplayRole or role == Qt.EditRole:
            hd = self.ogtGroup.heading_by_index(row)

            if col == self.C.head_code:
                return hd.head_code

            if col == self.C.unit:
                return hd.unit_label

            if col == self.C.type:
                return hd.type_label

            if col == self.C.head_description:
                return hd.head_description

            if col == self.C.valid:
                return "TODO"

            return "?"

        if role == Qt.DecorationRole:
            if col == self.C.head_code:
                return Ico.icon(Ico.AgsHeading)

        if role == Qt.FontRole:
            if col == self.C.head_code:
                f = QtWidgets.QFont()
                f.setBold(True)
                return f

        if role == Qt.TextAlignmentRole:
            if col == 0:
                return Qt.AlignRight
            if col in [self.C.valid, self.C.unit, self.C.type]:
                return Qt.AlignCenter
            return Qt.AlignLeft

        if False and role == Qt.BackgroundColorRole:
            cell = self.ogtDoc.group_by_index(row)[col]
            #bg = cell.get_bg()
            if len(self.ogtGroup.data_cell(row, col).errors) > 0:
                pass
            return QtWidgets.QColor(bg)


        return QtCore.QVariant()


    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            return QtCore.QVariant(self._col_labels[idx])

        if role == Qt.TextAlignmentRole and orient == Qt.Horizontal:
            if  idx == 0:
                return Qt.AlignRight
            if idx in [self.C.valid, self.C.unit, self.C.type]:
                return Qt.AlignCenter
            return Qt.AlignLeft

        return QtCore.QVariant()

class HeadersListWidget( QtWidgets.QWidget ):


    #sigGoto = pyqtSignal(str)

    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.ogtGroup = None

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)
        self.mainLayout.addWidget(xwidgets.XLabel("Foooo"))

        self.tree = QtWidgets.QTreeView()
        self.tree.setRootIsDecorated(False)
        self.mainLayout.addWidget(self.tree, 0)


        self.model = HeadersListModel()
        self.tree.setModel(self.model)

        self.tree.setColumnWidth(HeadersListModel.C.head_code, 110)
        self.tree.setColumnWidth(HeadersListModel.C.valid, 50)
        self.tree.setColumnWidth(HeadersListModel.C.unit, 50)
        self.tree.setColumnWidth(HeadersListModel.C.type, 50)


    def set_group(self, ogtGrp):
        self.model.set_group(ogtGrp)


class HeadCodeWidget( QtWidgets.QWidget ):
    """The info in tableWidget """

    #sigGoto = pyqtSignal(str)

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogtDoc
        self.ogtHeading = None

        self.debug = False
        self.setAutoFillBackground(True)

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)


        ## So splits up tthe header into parts..
        #self.headerWidget = QtWidgets.QWidget()
        self.headerCodeLay = xwidgets.hlayout()
        #self.headerWidget.setLayout(self.headerGridLay)
        self.mainLayout.addLayout(self.headerCodeLay)

        self.lblHeadCode = xwidgets.XLabel("-", bold=True)
        self.headerCodeLay.addWidget(self.lblHeadCode, 10)

        #self.buttGroup = xwidgets.XToolButton(self, text="group")
        #self.headerGridLay.addWidget(self.buttGroup)

        self.buttHeadCode = xwidgets.XToolButton(self, ico=Ico.BulletDown,  bold=True, popup=True, menu=True)
        self.buttHeadCode.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.headerCodeLay.addWidget(self.buttHeadCode, 0)

        self.buttHeadCode.menu().addAction("Open Group TODO")
        self.buttHeadCode.menu().addAction("Select another heading TODO")

        #sp = self.buttHeadCode.sizePolicy()
        #sp.setHorizontalPolicy(QtWidgets.QSizePolicy.Expanding)
        #self.buttHeadCode.setSizePolicy(sp)
        """
        sty = "background-color: #dddddd; color: black; padding: 3px; font-size: 8pt;"

        # description
        self.lblHeadDescription = QtWidgets.QLabel()
        self.lblHeadDescription.setStyleSheet(sty)
        self.lblHeadDescription.setFixedHeight(60)
        self.lblHeadDescription.setWordWrap(True)
        self.lblHeadDescription.setAlignment(Qt.AlignTop|Qt.AlignLeft)
        self.mainLayout.addWidget(self.lblHeadDescription)
        """

    def set_head_code(self, x):
        self.lblHeadCode.setText(x)


class TableHeaderWidget( QtWidgets.QWidget ):
    """The HEADER info in tableWidget """

    sigGoto = pyqtSignal(str)

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogtDoc
        self.ogtHeading = None

        self.debug = False


        self.mainLayout = QtWidgets.QGridLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        row = 0
        ## So splits up tthe header into parts..
        #self.headerWidget = QtWidgets.QWidget()
        self.headerGridLay = QtWidgets.QHBoxLayout()
        self.mainLayout.addLayout(self.headerGridLay, row, 0, 1, 3)

        self.lblHeadCode = xwidgets.XLabel("-", bold=True)
        self.headerGridLay.addWidget(self.lblHeadCode, 10)


        self.buttHeadCode = xwidgets.XToolButton(self, ico=Ico.BulletDown,  bold=True, popup=True, menu=True)
        self.buttHeadCode.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.headerGridLay.addWidget(self.buttHeadCode, 0)

        self.buttHeadCode.menu().addAction("Open Group TODO")
        self.buttHeadCode.menu().addAction("Select another heading TODO")


        sty = "background-color: #dddddd; color: black; padding: 3px; font-size: 8pt;"

        # description
        row += 1
        self.lblHeadDescription = QtWidgets.QLabel()
        self.lblHeadDescription.setStyleSheet(sty)
        self.lblHeadDescription.setFixedHeight(60)
        self.lblHeadDescription.setWordWrap(True)
        self.lblHeadDescription.setAlignment(Qt.AlignTop|Qt.AlignLeft)
        self.mainLayout.addWidget(self.lblHeadDescription, row, 0, 1, 3)


        # unit
        row += 1
        lbl = xwidgets.XLabel("Unit:", align=Qt.AlignRight, style=sty)
        self.mainLayout.addWidget(lbl, row, 0)

        self.lblUnit = xwidgets.XLabel("-", bold=True, style=sty + "color: #000099;")
        self.mainLayout.addWidget(self.lblUnit, row, 1, 1, 2)

        # Type
        row += 1
        lbl = xwidgets.XLabel("Type:", align=Qt.AlignRight, style=sty )
        self.mainLayout.addWidget(lbl, row, 0)

        self.lblType = xwidgets.XLabel("-", bold=True, style=sty + "color: #000099;" )
        self.mainLayout.addWidget(self.lblType, row, 1)

        self.buttLink = QtWidgets.QToolButton()
        #self.buttLink.setAutoRaise(True)
        self.buttLink.setText("Goto")
        self.mainLayout.addWidget(self.buttLink, row, 2)
        self.buttLink.setVisible(False)
        self.buttLink.clicked.connect(self.on_goto)

        self.mainLayout.setColumnStretch(0, 1)
        self.mainLayout.setColumnStretch(1, 5)

    def set_link(self, state):
        self.buttLink.setVisible(state)

    def set_heading(self, ogtHeading):

        self.ogtHeading = ogtHeading

        descr = self.ogtHeading.head_description
        t = "-" if descr == None else descr
        para = '<p style="line-height: 80%">' + t + '</p>'
        self.lblHeadDescription.setText(para)

        self.lblHeadCode.setText(ogtHeading.head_code)

        self.lblUnit.setText(self.ogtHeading.unit_label)
        #typ = "<a href="""
        self.lblType.setText(self.ogtHeading.type_label)
        #self.lblType.setToolTip(hrec['type'])

        typ = ags4.AGS4.type(self.ogtHeading.type)
        if typ:
            self.lblType.setToolTip(typ['description'])
        else:
            self.lblType.setToolTip(self.ogtHeading.type_label)

    def on_goto(self):
        self.sigGoto.emit(self.ogtHeading.head_code)




class HeadingsModel(QtCore.QAbstractTableModel):
    """Model for group headings"""
    sigChanged = pyqtSignal()

    class R:
        head_code = 0
        head_description = 1
        unit = 2
        data_type = 3
        _row_count = 4

    def __init__(self, parent=None, ogtGroup=None):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.ogtGroup = ogtGroup


    def columnCount(self, parent=None):
        return self.ogtGroup.headings_count

    def rowCount(self, parent=None):
        return self.R._row_count


    def data(self, index, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = index.row()
        col = index.column()
        head = self.ogtGroup.heading_from_index(col)

        if role == Qt.DisplayRole:

            if row == self.R.head_code:
                return None # head.head_code #- now a widget

            if row == self.R.head_description:
                return head.head_description

            if row == self.R.unit:
                return head.unit_cell.value

            if row == self.R.data_type:
                return head.data_type_cell.value

            return "_NOCELL_"

        if role == Qt.BackgroundColorRole:
            #cell = self.ogtGroup.raw_cell(row, col)
            if row == self.R.head_code:
                return QtGui.QColor(head.head_cell.background())

            if row == self.R.unit:
                return QtGui.QColor(head.unit_cell.background())

            if row == self.R.data_type:
                return QtGui.QColor(head.data_type_cell.background())

            return QtGui.QColor("white")


        if role == Qt.TextAlignmentRole:
            if row == self.R.head_description:
                return Qt.AlignTop

        return None

    def headerData(self, idx, orientation, role=Qt.DisplayRole):
        """Returns the headers to display"""

        if role == Qt.DisplayRole and orientation == Qt.Horizontal:

            return "" #self.ogtGroup.heading_from_index(idx).head_code

        if role == Qt.DisplayRole and orientation == Qt.Vertical:

            if idx == self.R.head_code:
                return "Code"

            if idx == self.R.head_description:
                return "Desc"

            if idx == self.R.unit:
                return "Unit"

            if idx == self.R.data_type:
                return "Type"

            return " "

        if role == Qt.TextAlignmentRole and orientation == Qt.Vertical:
                return Qt.AlignRight

        return None

    def setData(self, index, value, role=None):
        """Updates DATA when modified in the view"""
        if role == Qt.EditRole:
            success = self.ogtGroup.set_data_cell_value(index.row(), index.column(), value)
            self.layoutChanged.emit()
            return success
            #self.dataChanged.emit(index, index)
            #return True
        return False

    def flags(self, index):
        """Returns the flags of the model"""
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled # | Qt.ItemIsEditable
        return flags


class DataModel(QtCore.QAbstractTableModel):
    """Model for group data
    """

    #sigChanged = pyqtSignal()

    def __init__(self, parent, ogtGroup):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.ogtGroup = ogtGroup

    def columnCount(self, parent=None):
        return self.ogtGroup.headings_count

    def rowCount(self, parent=None):
        return self.ogtGroup.data_rows_count


    def data(self, index, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = index.row()
        col = index.column()

        if role == Qt.DisplayRole or role == Qt.EditRole:
            cell = self.ogtGroup.data_cell(row, col)
            if cell:
                return cell.value
            return None

        if role == Qt.BackgroundColorRole:
            cell = self.ogtGroup.data_cell(row, col)
            if cell:
                return QtGui.QColor(cell.background())
            return QtGui.QColor("#dddddd")

        if False and role == Qt.BackgroundColorRole:
            cell = self.ogtGroup.data_cell(row, col)
            bg = cell.get_bg()
            if len(self.ogtGroup.data_cell(row, col).errors) > 0:
                pass
            return QtGui.QColor(bg)



    def headerData(self, idx, orientation, role=Qt.DisplayRole):

        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.ogtGroup.heading_from_index(idx).head_code

        if role == Qt.DisplayRole and orientation == Qt.Vertical:
            return str(idx + 1)

        if role == Qt.TextAlignmentRole and orientation == Qt.Vertical:
                return Qt.AlignRight

        return None

    def setData(self, index, value, role=None):
        """Updates DATA when modified in the view"""
        if role == Qt.EditRole:
            cell = self.ogtGroup.data_cell(index.row(), index.column())
            cell.value = value.strip()
            self.ogtGroup.emit()
            #success = self.ogtGroup.set_data_cell_value(index.row(), index.column(), value)
            #self.layoutChanged.emit()
            #self.modelReset.emit()
            #self.sigChanged.emit()
            return True
            #self.dataChanged.emit(index, index)
            #return True
        return False

    def flags(self, idx):

        head = self.ogtGroup.heading_from_index(idx.column())
        dt = head.data_type
        if dt == "PA": # or dt == "ID"
            flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        else:
            flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable
        #if index.column() != 0:
        #    flags |= Qt.ItemIsEditable
        return flags

    def midx_from_val(self, cidx, v):
        for ridx in range(0, self.ogtGroup.data_rows_count):
            dcell = self.ogtGroup.data_cell(ridx, cidx)
            if dcell and dcell.value == v:
                return self.index(ridx, cidx)
        return None

    def val_from_midx(self, midx):
        return self.ogtGroup.data_cell( midx.row(), midx.column() )

    def midx_from_cell(self, srcCell):
        for ridx in range(0, self.ogtGroup.data_rows_count):
            cells = self.ogtGroup.data[ridx]
            for ki, cell in cells.items():
                if cell == srcCell:
                    return self.index(ridx, self.ogtGroup.index_of_heading(ki))
        return None



T_STYLE = """
QTableView
{
       
}
QTableView::item
{
  
}
QTableView::item:selected {
    border: 1px solid black;
}

"""
class XTableKeysHeadingView(QtWidgets.QTableView):
    """Catches the Down button in the headings view, fires sigDown to move to table below"""
    sigMoveDown = pyqtSignal(int)

    def __init__(self, parent=None):
        QtWidgets.QTableView.__init__(self, parent)


    def keyPressEvent(self,  ev):
        indexes = self.selectionModel().selectedIndexes()
        if len(indexes) == 1 and ev.key() == Qt.Key_Down and indexes[0].row() == 3:
            self.sigMoveDown.emit(indexes[0].column())
            return
        QtWidgets.QTableView.keyPressEvent(self, ev)


class XTableKeysDataView(QtWidgets.QTableView):
    """Catched the UP button in the data view, fires sigUp to move to table above"""
    sigMoveUp = pyqtSignal(int)

    def __init__(self, parent=None):
        QtWidgets.QTableView.__init__(self, parent)

    def keyPressEvent(self,  ev):
        indexes = self.selectionModel().selectedIndexes()
        if len(indexes) == 1 and ev.key() == Qt.Key_Up and indexes[0].row() == 0:
            self.sigMoveUp.emit(indexes[0].column())
            return
        QtWidgets.QTableView.keyPressEvent(self, ev)





class GroupTableWidget( QtWidgets.QWidget ):
    """The Group Table..... consists of two tables.. and two models
        - the headings model + tableHeaders
        - the data model + tableData
    """

    sigGoto = pyqtSignal(str)


    def __init__( self, parent=None, ogtGroup=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtGroup = ogtGroup


        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        #== Headings table and model ====
        self.modelHeadings = HeadingsModel(ogtGroup=ogtGroup)

        self.tableHeadings = XTableKeysHeadingView()
        self.mainLayout.addWidget(self.tableHeadings, 0)
        self.tableHeadings.setModel(self.modelHeadings)

        self.tableHeadings.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        self.tableHeadings.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableHeadings.horizontalHeader().setFixedHeight(15)
        self.tableHeadings.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.tableHeadings.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        self.tableHeadings.doubleClicked.connect(self.on_table_headings_dbl_clicked)
        self.tableHeadings.selectionModel().selectionChanged.connect(self.on_table_headings_selection_changed)
        self.tableHeadings.sigMoveDown.connect(self.on_table_headings_move_down)
        self.tableHeadings.horizontalScrollBar().valueChanged.connect(self.on_table_headings_h_scroll)
        self.tableHeadings.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tableHeadings.customContextMenuRequested.connect(self.on_table_headings_context)
        self.tableHeadings.horizontalHeader().sectionResized.connect(self.on_col_resize)

        #= Popup menu for tableheader
        self.popHead = QtWidgets.QMenu()
        wi = QtWidgets.QWidgetAction(self.popHead)
        self.popHeadLbl = QtWidgets.QLabel()
        self.popHeadLbl.setText("-")
        self.popHeadLbl.setStyleSheet("background-color: #666666; color: #eeeeee; padding: 4px;")
        wi.setDefaultWidget(self.popHeadLbl)
        self.popHead.addAction(wi)

        #self.popHead.addAction(Ico.icon(Ico.Add), "Add Heading...", self.on_add_heading)
        self.popHead.addSeparator()
        self.popHead.addAction(Ico.icon(Ico.Ags4), "Open AGS spec")



        #== Data
        self.modelData =  DataModel(self, ogtGroup)


        self.tableData = XTableKeysDataView()
        self.mainLayout.addWidget(self.tableData, 200)#
        self.tableData.setStyleSheet(T_STYLE)


        self.tableData.setModel(self.modelData)

        self.tableData.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        self.tableData.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableData.horizontalHeader().hide()
        self.tableData.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        self.tableData.horizontalScrollBar().valueChanged.connect(self.on_table_data_h_scroll)

        self.tableData.verticalHeader().setFixedWidth(self.tableHeadings.verticalHeader().width())

        self.tableData.doubleClicked.connect(self.on_table_data_dbl_clicked)

        self.tableData.selectionModel().selectionChanged.connect(self.on_table_data_selection_changed)
        self.tableData.sigMoveUp.connect(self.on_table_data_move_up)



        self.set_heading_height()

    def do_update(self):
        self.modelHeadings.modelReset.emit()
        self.modelData.modelReset.emit()

        self.update_table_heading_widgets()

    def update_table_heading_widgets(self):

        for cidx in range(0, self.modelHeadings.columnCount()):
            # first row is HEAD_CODE
            hidx = self.modelHeadings.index(0, cidx)
            widget = self.tableHeadings.indexWidget(hidx)
            oHead = self.ogtGroup.heading_from_index(cidx)
            if widget == None:

                widget = HeaderCellWidget(ogtHead=oHead)
                self.tableHeadings.setIndexWidget(hidx, widget)
            #head_code = self.ogtGroup.headings_index[cidx]
            #widget.set_label(head_code)
            widget.parentheading = oHead
            widget.do_update()


    def set_heading_height(self):
        HEADER_HEIGHT = 134
        DESC_HEIGHT = 60
        self.tableHeadings.setRowHeight(1, DESC_HEIGHT)

        h = HEADER_HEIGHT + 35
        if self.tableHeadings.isRowHidden(1):
            h = h - DESC_HEIGHT
        self.tableHeadings.setFixedHeight(h)


    def on_col_resize(self, idx):
        self.tableData.setColumnWidth(idx, self.tableHeadings.columnWidth(idx))


    def on_table_headings_context(self, qPoint):

        clickIdx = self.tableHeadings.indexAt(qPoint)
        #headcodeIdx = self.modelHeadings.index(0, clickIdx.column())

        head_code = self.ogtGroup.heading_from_index(clickIdx.column()).head_code
        self.popHeadLbl.setText(head_code)

        self.popHead.exec_(self.tableHeadings.mapToGlobal(qPoint) )

    def on_table_headings_dbl_clicked(self, midx):

        row = midx.row()
        col = midx.column()
        R = HeadingsModel.R

        if row == R.unit:
            v = self.ogtGroup.heading_from_index(col)
            if v:
                dial = ogtgui_dialogs.UnitSelectDialog(self, v)
                dial.exec_()


        if row == R.data_type:
            ogtHead = self.ogtGroup.heading_from_index(col)
            if ogtHead:
                dial = ogtgui_dialogs.DataTypeSelectDialog(self, ogtHead)
                dial.exec_()



    def on_table_data_dbl_clicked(self, midx):
        row = midx.row()
        col = midx.column()

        headOb = self.ogtGroup.heading_from_index(col)

        if headOb.data_type == "PA":
            ## Its a picklist
            cell = self.ogtGroup.data_cell(row, col)
            dial = ogtgui_dialogs.AbbrPickDialog(self, ogtHead=headOb, editCell=cell)
            resp = dial.exec_()
            return


    def on_table_data_h_scroll(self, x):
        self.tableHeadings.horizontalScrollBar().setValue(x)

    def on_table_headings_h_scroll(self, x):
        self.tableData.horizontalScrollBar().setValue(x)


    def on_goto(self, code):
        self.sigGoto.emit(code)

    def set_show_description(self, state):
        self.tableHeadings.setRowHidden(1, not state)
        self.set_heading_height()

    def select_error(self, ogtErr):

        midx = self.modelData.midx_from_cell(ogtErr.cell)
        if midx:
            self.tableData.selectionModel().select(midx, QtCore.QItemSelectionModel.SelectCurrent)
            self.tableData.scrollTo(midx)


    def on_table_headings_selection_changed(self, sel, desel):
        if sel.count() == self.tableData.selectionModel().hasSelection():
            self.tableData.blockSignals(True)
            self.tableData.selectionModel().clearSelection()
            self.tableData.blockSignals(False)

    def on_table_data_selection_changed(self, sel, desel):
        if sel.count() == self.tableHeadings.selectionModel().hasSelection():
            self.tableHeadings.blockSignals(True)
            self.tableHeadings.selectionModel().clearSelection()
            self.tableHeadings.blockSignals(False)


    def on_table_data_move_up(self, cidx):
        self.tableData.blockSignals(True)
        self.tableData.selectionModel().clearSelection()
        self.tableData.blockSignals(False)
        midx = self.modelHeadings.index(3, cidx)
        self.tableHeadings.selectionModel().select(midx, QtCore.QItemSelectionModel.SelectCurrent)
        self.tableHeadings.setCurrentIndex(midx)
        self.tableHeadings.setFocus()

    def on_table_headings_move_down(self, cidx):
        self.tableHeadings.blockSignals(True)
        self.tableHeadings.selectionModel().clearSelection()
        self.tableHeadings.blockSignals(False)
        midx = self.modelData.index(0, cidx)
        self.tableData.selectionModel().select(midx, QtWidgets.QItemSelectionModel.SelectCurrent)
        self.tableData.setCurrentIndex(midx)
        self.tableData.setFocus()


class GroupViewWidget( QtWidgets.QWidget ):

    sigGoto = pyqtSignal(str)
    #sigChanged = pyqtSignal()

    def __init__( self, parent=None, ogtGroup=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtGroup = ogtGroup

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        ##=== title and toolbar at top =========
        m = 3
        topLay = xwidgets.hlayout(margin=m, spacing=1)
        self.mainLayout.addLayout(topLay, 0)

        #== description
        sty = " color: #444444; padding: 2px; font-size: 14pt;"
        self.lblGroupDescription = QtWidgets.QLabel()
        self.lblGroupDescription.setText("-- empty --")
        self.lblGroupDescription.setStyleSheet(sty + "")
        topLay.addWidget(self.lblGroupDescription, 100)

        # == Actions
        self.tbgActions = xwidgets.ToolBarGroup(title="Actions", width=120)
        topLay.addWidget(self.tbgActions, 0)
        # The AGS group data
        self.buttAddHeading = xwidgets.XToolButton(text="Add Heading", ico=Ico.Add, callback=self.on_add_heading)
        self.tbgActions.addWidget(self.buttAddHeading)
       # self.buttAddHeading.clicked.connect(self.on_butt_ags_group_code)


        # == Group Info button
        self.tbgInfo = xwidgets.ToolBarGroup(title="Info", width=80)
        topLay.addWidget(self.tbgInfo, 0)
        # The AGS group data
        self.buttAgsGroupCode = xwidgets.XToolButton(text="-", ico=Ico.Ags4, tooltip="View AGS data Dict")
        self.tbgInfo.addWidget(self.buttAgsGroupCode)
        self.buttAgsGroupCode.clicked.connect(self.on_butt_ags_group_code)


        #== Parent + kids ----
        ## button group for convenience in disabling etc
        self.buttGroupReleations = QtWidgets.QButtonGroup()
        self.buttGroupReleations.setExclusive(False)

        #= Parent
        self.tbgParent = xwidgets.ToolBarGroup(title="Parent", width=80  )
        topLay.addWidget(self.tbgParent, 0)
        self.buttParent = xwidgets.XToolButton(text="", ico=Ico.BulletBlack)
        self.tbgParent.addWidget(self.buttParent)

        #= Children
        self.tbgChildren = xwidgets.ToolBarGroup(title="Children")
        topLay.addWidget(self.tbgChildren, 0)
        for i in range(0, 2):
            butt = xwidgets.XToolButton(text="", ico=Ico.BulletBlack)
            butt.setFixedWidth(60)
            self.tbgChildren.addWidget(butt)
            self.buttGroupReleations.addButton(butt, 0)

        self.buttGroupReleations.addButton(self.buttParent, 2) # adding parent at end for access
        self.buttGroupReleations.buttonClicked.connect(self.on_butt_parent_child)


        #== View Mode
        self.tbgView = xwidgets.ToolBarGroup(title="View", is_group=True, toggle_icons=True, toggle_callback=self.on_view_change )
        topLay.addWidget(self.tbgView)

        self.tbgView.addButton(text="Data", idx=0, checkable=True)
        self.tbgView.addButton(text="Source", idx=1, checkable=True, checked=True)





        # ============ mid splitter with stack widget
        self.splitter = QtWidgets.QSplitter()
        self.mainLayout.addWidget(self.splitter, 100)


        #== Left layout
        self.leftWidget = QtWidgets.QWidget()
        self.leftLay = xwidgets.vlayout()
        self.leftWidget.setLayout(self.leftLay)
        self.splitter.addWidget(self.leftWidget)

        self.groupDataTableWidget = GroupTableWidget(ogtGroup=self.ogtGroup)
        self.leftLay.addWidget(self.groupDataTableWidget)

        #self.sourceCellsTableWidget = SourceCellsTableWidget(self, self.ogtGroup)
        #self.leftLay.addWidget(self.sourceCellsTableWidget)
        #self.sourceCellsTableWidget.sigSelect.connect(self.on_source_table_select)

        #== Right layout
        self.rightWidget = QtWidgets.QWidget()
        self.rightLay = xwidgets.vlayout()
        self.rightWidget.setLayout(self.rightLay)
        self.splitter.addWidget(self.rightWidget)

        self.errorsWidget = ogtgui_errors.ErrorsWidget(self, ogtgui_errors.VIEW_ERR_MODE.group, ogtGroup=self.ogtGroup)
        self.rightLay.addWidget(self.errorsWidget, 1)
        self.errorsWidget.sigGotoError.connect(self.on_goto_error)


        self.headersListWidget = HeadersListWidget()
        self.headersListWidget.setMinimumWidth(300)
        self.rightLay.addWidget(self.headersListWidget, 1)



        self.splitter.setStretchFactor(0, 10)
        self.splitter.setStretchFactor(1, 4)
        # TODO restore
        #G.settings.restore_splitter(self.splitter)


        self.groupDataTableWidget.modelHeadings.sigChanged.connect(self.on_changed)
        #self.groupDataTableWidget.modelData.sigChanged.connect(self.on_changed)

        self.ogtGroup.parentDoc.sig_changed.connect(self.on_data_changed)
        #self.on_data_changed()

    def on_butt_parent_child(self, butt):
        group_code = str(butt.text())
        self.sigGoto.emit(group_code)

    def set_show_description(self, state):
        self.groupDataTableWidget.set_show_description(state)

    def on_source_table_select(self, cell):
        self.errorsWidget.select_from_cell(cell)

    def on_headings_data_changed(self):
        print( "on_headings_data_changed", self)

    def on_changed(self):
        self.sigChanged.emit()


    def do_update(self):
        """Updates all widgets and tables etc"""


        for model in [  #self.groupDataTableWidget.headingsModel,
                        #self.groupDataTableWidget.dataModel,
                        self.errorsWidget.model]:
            model.modelReset.emit()

        self.groupDataTableWidget.do_update()

        #==================================
        #= update toolbar labels and butts
        self.lblGroupDescription.setText(self.ogtGroup.group_description)
        self.buttAgsGroupCode.setText(self.ogtGroup.group_code)

        self.clear_relation_buttons()
        if self.ogtGroup.parent:
            self.set_relation_butt(self.buttParent, self.ogtGroup.parent)
        if self.ogtGroup.children:
            for idx, kid in enumerate(self.ogtGroup.children):
                butt = self.buttGroupReleations.button(idx)
                self.set_relation_butt(butt, kid)

    def clear_relation_buttons(self):
        for butt in self.buttGroupReleations.buttons():
            butt.setIcon(Ico.icon(Ico.BulletBlack))
            butt.setText("")
            butt.setDisabled(True)

    def set_relation_butt(self, butt, grp_code):
        butt.setIcon(Ico.icon(Ico.AgsGroup))
        butt.setText(grp_code)
        butt.setEnabled(True)

    def on_data_changed(self):

        #return
        #rub()
        for model in [
                    #self.groupSourceTableWidget.model,
                    self.groupDataTableWidget.headingsModel,
                    self.groupDataTableWidget.dataModel    ]:
            if model == self.sender():
                pass
            else:
                model.modelReset.emit()

        self.headersListWidget.model.layoutChanged.emit()
        self.errorsWidget.model.layoutChanged.emit()

        self.groupDataTableWidget.update_headings()


    def deadupdate_group(self, ogtGroup):

        self.ogtGroup = ogtGroup
        ## Set the labels
    def deadupdate_widgets(self):
        deaddd()
        self.buttGroupCode.setText( self.ogtGroup.group_code  )
        self.lblGroupDescription.setText( "-" if self.ogtGroup.group_description == None else self.ogtGroup.group_description )

        return
        # load into widgets

        self.groupDataTableWidget.set_group(ogtGroup)
        self.groupSourceTableWidget.set_group(ogtGroup)
        self.headersListWidget.set_group(ogtGroup)
        self.errorsWidget.set_group(ogtGroup)
        return


    def deadon_goto(self, code):
        self.sigGoto.emit(code)

    def on_goto_error(self, ogtErr):

        #self.sourceCellsTableWidget.select(ogtErr)
        self.groupDataTableWidget.select_error(ogtErr)

    def on_butt_ags_group_code(self):
        ags4_widgets.show_group_dialog(self, group_code=self.ogtGroup.group_code)


    def on_view_change(self, idx):
        return
        self.stackWidget.setCurrentIndex(idx)

    def on_add_heading(self):
        d = ogtgui_dialogs.HeadingsSelectDialog(self, self.ogtGroup)
        d.exec_()


class HeaderCellWidget(QtWidgets.QWidget):

    def __init__(self, parent=None, ogtHead=None):
        QtWidgets.QWidget.__init__(self, parent)

        self.ogtHead = ogtHead

        m = 1
        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(m,m,m,m)
        self.setLayout(self.mainLayout)

        self.lblCode = xwidgets.XLabel()
        self.mainLayout.addWidget(self.lblCode, 10)


        ww = 10
        self.lblKey = xwidgets.XLabel(width=ww, align=Qt.AlignCenter, text="K", style="background-color: yellow;")
        self.mainLayout.addWidget(self.lblKey, 0)
        self.lblKey.setToolTip("Key Field")

        self.lblReq = xwidgets.XLabel(width=ww, align=Qt.AlignCenter, text="R", style="background-color: #F0DAED;")
        self.mainLayout.addWidget(self.lblReq, 0)
        self.lblReq.setToolTip("Required")


    def set_label(self, head_code):
        self.lblCode.setText(head_code)

        #self.do_update()

    def do_update(self):

        self.lblCode.setText(self.ogtHead.head_code)

        self.lblKey.setVisible(self.ogtHead.key == True)
        self.lblReq.setVisible(self.ogtHead.required == True)




# coding: utf-8

from ogt.ogt_heading import OGTHeading
from ogt import ags4
from ogt import ogt_validate
from ogt import ogt_cell


class OGTGroup:
    """Represents :term:`GROUP` with its headings and data"""

    def __init__(self, parentDoc=None,  group_code_cell=None, src_ref=None):
        """
        :param parentDoc: The document object this group is within
        :type parentDoc: OGTDocument
        :param group_code_cell: The cell containing the group code
        :type group_code_cell: OGTCell
        """

        assert parentDoc != None
        self.parentDoc = parentDoc
        """Pointer to parent document instance

        :type: OGTDocument
        """

        self.group_code_cell = group_code_cell
        """The four character group code"""

        self.headings_index = []
        self.headings_dict = {}
        """A `list` of head_code > ogtHeadings"""

        self.lost_idxs = []

        self._errors_cache = None

        self.data = []

    def __repr__(self):
        return "<Group '%s'>" % self.group_code

    @property
    def group_code(self):
        return self.group_code_cell.value

    @property
    def group_description(self):
        if self.data_dict():
            return self.data_dict()['group_description']
        return None

    @property
    def group_required(self):
        if self.data_dict():
            return bool(self.data_dict()['group_required'])
        return None

    @property
    def parent(self):
        if self.data_dict():
            return self.data_dict()['parent']
        return None

    @property
    def children(self):
        if self.data_dict():
            return self.data_dict()['children']
        return None

    def add_get_headings_row(self, row_cells, src_ref=None):

        # TODO src_ref

        ret_idx = []

        for cidx, hcell in enumerate(row_cells):
            # validate the headings string
            head_code = hcell.value
            # check heading not exists already
            oHead = self.heading_from_code(head_code)
            if oHead is None:
                # add to dic
                ogtHead = OGTHeading(self, hcell)
                hcell.parentHeading = ogtHead
                self.headings_dict[head_code] = ogtHead

                # TODO this needs to be inserted in correct plage
                self.headings_index.append(head_code)
            # add to list of headings used in the group/import
            ret_idx.append(head_code)

        return ret_idx

    def add_heading(self, rec):

        # check its there already
        head_code = rec['head_code']
        ogtHead = self.heading_from_code(head_code)
        if ogtHead:
            return ogtHead

        # get a list of dd heading for ordering
        head_order = ags4.DD.group(self.group_code)['headings_index']

        # get current heading
        curr_heads = list(self.headings_index)
        curr_heads.append(head_code)

        # make required sort order
        req_order = []
        for gc in head_order:
            if gc in curr_heads:
                req_order.append(gc)
                curr_heads.remove(gc)
        req_order.extend(curr_heads)

        insert_idx = req_order.index(head_code)

        hcell = ogt_cell.OGTCell(raw=head_code)
        ogtHead = OGTHeading(self, hcell)
        hcell.parentHeading = ogtHead
        self.headings_index.insert(insert_idx, head_code)
        self.headings_dict[head_code] = ogtHead

        print(self.headings_index)

        unitCell = ogt_cell.OGTCell(raw=rec['unit'], parent_heading=ogtHead)
        ogtHead.set_unit(unitCell)

        dataTypeCell = ogt_cell.OGTCell(raw=rec['data_type'], parent_heading=ogtHead)
        ogtHead.set_data_type(dataTypeCell)

        for dic in self.data:
            dic[head_code] = ogt_cell.OGTCell(raw="", parent_heading=ogtHead)

        self.emit()
        return ogtHead

    def heading_from_code(self, head_code):
        return self.headings_dict.get(head_code)

    def deadhas_heading(self, head_code):
        return head_code in self.headings_dict

    # def headings_sort(self):
    #     return self.headings_index
    #     # TODO
    #     if self._headings_sort is None:
    #         dd = self.data_dict()
    #         if dd is None:
    #             return []
    #         self._headings_sort = []
    #         if dd.headings_sort() is None:
    #             return self.headings.keys()
    #         else:
    #             for head_code in dd.headings_sort():
    #                 if head_code in self.headings:
    #                     self._headings_sort.append(head_code)

        return self._headings_sort

    @property
    def headings_list(self):
        """Return a list of heading """
        lst = []
        for hcode in self.headings_index:
            lst.append(self.headings_dict[hcode])
        return lst

    def heading_from_index(self, idx):
        # print(self.headings_index, idx)
        return self.headings_dict.get(self.headings_index[idx])

    def index_of_heading(self, head_code):
        for cidx, hc in enumerate(self.headings_index):
            if hc == head_code:
                return cidx
        return None

    @property
    def headings_count(self):
        """ no of headings in this group """
        return len(self.headings_index)

    def add_data_row(self, row_cells):
        self.data.append(row_cells)

    def insert_data_row(self, ridx, row_cells):
        self.data.insert(ridx, row_cells)

    def data_cell(self, ridx, cidx):
        """Return cells from row col, if it exists"""
        return self.data[ridx].get(self.headings_index[cidx])

    def raw_cell(self, ridx, cidx):
        try:
            return self._raw_rows[ridx][cidx]
        except IndexError:
            return None
        return ARGHH

    def add_lost_row(self, row_cells, lidx):
        """Not sure, but we need add rows with no descriptors, or merge erors"""
        self.lost_rows.append(lidx)

    def data_row_dict(self, ridx):
        dic = {}
        for hc, cell in self.data[ridx].items():
            dic[hc] = cell.value
        return dic

    def data_rows_dict(self):
        lst = []
        for row in self.data:
            dic = {}
            for hc, cell in row.items():
                dic[hc] = cell.value
            lst.append(dic)
        return lst

    def data_column_from_head_code(self, head_code, as_cells=True):
        lst = []
        for ridx in range(0, self.data_rows_count):
            cell = self.data[ridx].get(head_code)
            if as_cells:
                lst.append(cell)
            else:
                if cell is not None:
                    lst.append(cell.value if cell else None)
        return lst

    @property
    def data_rows_count(self):
        return len(self.data)

    def data_dict(self):
        """Returns the data dictionary for this group, if it exists

        :return: dict if exists, else `None`

        """

        # TODO maybe this needs to returen custom stuff as well ?

        return ags4.DD.group(self.group_code)

    def to_dict(self):
        """
        :rtype: dict
        :return: A dictionary with the ags structures data (see :ref:`extended_mode`)
        """

        if self.parentDoc.opts.extended:

            # shortcut to data dict
            grp_dd = self.data_dict()
            # TODO cleanup data dict mess here

            heads = []
            for idx, head_code in enumerate(self.headings_index):

                ags_data_dict = None
                if head_code in grp_dd['headings_index']:
                    cidx = grp_dd['headings_index'].index(head_code)
                    ags_data_dict = grp_dd['headings'][cidx]

                heads.append({
                            "head_code": head_code,
                            "data_dict": ags_data_dict,
                            "unit": self.unit_for_heading(head_code),
                            "type": self.data_type_for_heading(head_code)
                            })

            dic = {
                    "group_code": self.group_code,
                    "data_dict": grp_dd,
                    "headings": heads,
                    "headings_sort": self.headings_index,
                    "data": self.data_rows_dict()
            }
            return dic

        dic = {
            ags4.DESCRIPTOR.heading: self.headings_index,
            ags4.DESCRIPTOR.unit: self.units_list(),
            ags4.DESCRIPTOR.type: self.data_types_list(),
            ags4.DESCRIPTOR.data: self.data_rows_dict()
        }

        return dic

    def add_units_row(self, row_cells):

        self.rows.append(row_cells)
        self._update_col_count(row_cells)

    def units_list(self):

        lst = []
        for hcode in self.headings_index:
            lst.append(self.headings_dict[hcode].unit)

        return lst

    def unit_for_heading(self, hcode):
        # TODO cleanup units
        if hcode in self.headings_dict:
            return self.headings_dict[hcode].unit
        return None  # should never happen

    def set_data_types_row(self, row_cells, lidx):

        self.types_idx = lidx - self.group_start_lidx
        self.rows.append(row_cells)
        self._update_col_count(row_cells)
        for idx, cell in enumerate(row_cells[1:]):
            self._headings[idx].set_type(cell)

    def data_types_list(self):

        lst = []
        for hcode in self.headings_index:
            lst.append(self.headings_dict[hcode].data_type)
        return lst

    def data_type_for_heading(self, hcode):
        return self.headings_dict[hcode].data_type

    @property
    def errors_warnings_count(self):
        if self._errors_cache is None:
            self._update_errors_cache()
        return len(self._errors_cache)

    def _update_errors_cache(self):
        self._errors_cache = []
        for ridx in range(0, self.data_rows_count):
            for cell in self.data[ridx].values():  # TODO
                self._errors_cache.extend(cell.errors_list())

    def errors_list(self):
        if self._errors_cache is None:
            self._update_errors_cache()
        return self._errors_cache

    def emit(self):
        self.parentDoc.emit()

    def clear_errors(self):
        self._errors_cache = None
        for ridx, dic in enumerate(self.data):

            for cell in dic.values():
                if isinstance(cell, ogt_cell.OGTCell):
                    cell.clear_errors()
                else:
                    pass

    def validate(self):

        # ogt_validate.group_code(self.raw_cell(0, 1))
        NOS = [str(n) for n in range(0, 10)]
        NOS.append(".")

        # headings code
        for headObj in self.headings_dict.values():

            ogt_validate.heading_code(self.group_code, headObj.head_cell)
            ogt_validate.data_type(headObj.data_type_cell)

            hcode = headObj.head_code

            # validate data
            if headObj.data_type == "PA":
                # check picklist
                # TODO
                pickcodes = self.parentDoc.abbrs(hcode, code_list=True)
                for ridx in range(0, self.data_rows_count):
                    cells_dic = self.data[ridx]
                    cell = cells_dic.get(hcode)
                    if cell:
                        if not cell.value:
                            continue
                        if pickcodes is not None and cell.value not in pickcodes:
                            cell.add_error(ogt_validate.OGTError(cell, "Invalid code", warn=False))

            if headObj.data_type in ["1DP", "2DP", "3DP", "4DP"]:
                # nos = NOS
                # nos.append(".")
                for ridx in range(0, self.data_rows_count):
                    cells_dic = self.data[ridx]
                    cell = cells_dic.get(hcode)
                    if cell:
                        if cell.value != "":
                            # validate nos and .
                            for c in cell.value:
                                if c not in NOS:
                                    cell.add_error(ogt_validate.OGTError(cell, "Not a NO", warn=False))
                                    break

                            # check decimals
                            dp = int(headObj.data_type[0:1])
                            parts = cell.value.split(".")
                            if len(parts) == 0 or len(parts[1]) != dp:
                                cell.add_error(ogt_validate.OGTError(cell, "Incorrect decimals", warn=False))

            if headObj.data_type == "IDDEADDDDD":
                # ID
                # ids = self.get_column(hcode)

                for ridx in range(0, self.data_rows_count):
                    cells_dic = self._data[ridx]
                    cell = cells_dic.get(hcode)
                    if cell:
                        if cell.value not in pickcodes:
                            cell.add_error(ogt_validate.OGTError(cell, "Invalid code", warn=False))

        return
        for ridx, row in enumerate(self._raw_rows):

            for cidx, cell in enumerate(row):

                # validate descriptor

                if ridx == self.headings_idx:
                    # validate headings
                    cell.value, fatal, errs = ags4.validate_heading_str(cell.value, cidx=cidx, lidx=ridx)
                    self.add_errors(errs)

                    if not fatal:
                        errs = ags4.validate_heading_ags(cell.value, self.group_code, cidx=cidx, lidx=ridx)

                        self.add_errors(errs)

                if ridx == self.types_idx:
                    pass



# -*- coding: utf-8 -*-


from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal

from ogt import ags4

#import app_globals as G
from ogtgui.settings import Settings

from ogtgui.img import Ico
from ogtgui import xwidgets
from ogtgui.ags4_models import AgsData, CG, AGS4_COLORS, HeadingsModel, AbbrevItemsModel
from ogtgui import ags4_widgets



def get_edit_widget(data_type, value=None):

    dt = data_type.upper().strip()

    ## PickList
    if dt == "PA":
        w = QtWidgets.QComboBox()
        w.addItem("-", "")
        return w

    ## Yes No
    if dt == "YN":
        w = QtWidgets.QComboBox()
        w.addItem("-", "")
        w.addItem("Yes", "1")
        w.addItem("No", "0")
        return w

    ## time
    if dt == "T":
        w = QtWidgets.QLineEdit()
        w.setInputMask("00:00:00")
        return w

    ## Deg, Mins, Secs
    if dt == "DMS":
        w = QtWidgets.QLineEdit()
        w.setInputMask("000.00.00")
        return w

    ## Date Time
    if dt == "DT":
        w = QtWidgets.QDateTimeEdit()
        w.setDisplayFormat("dd-mm-yyyy HH:mm:ss")
        return w

    if dt.endswith("DP"):

        opts = ["0DP", '1DP', "2DP", "3DP", "4DP"]
        dp = opts.index(dt) # get sigIndex from array position
        if dp == 0:
            # No decimals so a spinbox
            w = QtWidgets.QSpinBox()
            return w

        w = QtWidgets.QDoubleSpinBox()
        w.setDecimals(dp)
        return w

    if dt.endswith("SF"):
        w = QtWidgets.QDoubleSpinBox()
        w.setDecimals(5)
        return w

    if dt in ["X", "U", "XN"]:
        w = QtWidgets.QLineEdit()
        return w


    return None

def show_group_dialog(parent, group_code):
    d = AGS4GroupViewDialog(parent=parent, group_code=group_code)
    d.exec_()

class AGS4DataDictBrowser( QtWidgets.QWidget ):

    def __init__( self, parent=None, mode=None ):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False

        AgsData.load_init()

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        ##=============================================================
        self.tabWidget = QtWidgets.QTabWidget()
        self.mainLayout.addWidget(self.tabWidget)


        ##=============================================================
        self.agsGroupsWidget = AGS4GroupsBrowser(self)
        self.tabWidget.addTab(self.agsGroupsWidget, Ico.icon(Ico.AgsGroups), "Groups")

        self.unitsTypesWidget = AGS4UnitsTypesWidget(self)
        self.tabWidget.addTab(self.unitsTypesWidget, Ico.icon(Ico.AgsField), "Units && Types")

        ##=============================================================
        #self.agsAbbrevsWidget = AgsAbbrevsWidget.AgsAbbrevsWidget(self)
        #self.tabWidget.addTab(self.agsAbbrevsWidget,dIco.icon(dIco.AgsAbbrevs),  "Abbreviations / Pick lists")

        ##=============================================================
        #self.agsUnitsWidget = AgsUnitsWidget.AgsUnitsWidget(self)
        #self.tabWidget.addTab(self.agsUnitsWidget, dIco.icon(dIco.AgsUnits), "Units")



    def init_load(self):

        # load data dict
        #G.ags.init_load()
        #self.tabWidget.setCurrentIndex(1)

        self.agsGroupsWidget.set_focus()


class AGS4GroupsBrowser( QtWidgets.QWidget ):
    """The left panel with the classes, filter and groups table underneath"""

    sigGroupSelected = pyqtSignal(object)

    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.setObjectName("AGS4GroupsBrowser")

        self.proxy = QtCore.QSortFilterProxyModel()
        self.proxy.setSourceModel(AgsData.modelGroups)
        self.proxy.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)


        ##===============================================
        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setObjectName(self.objectName() + "groups_splitter")
        self.mainLayout.addWidget(self.splitter)

        ##############################################################################
        leftWidget = QtWidgets.QWidget()
        leftLayout = xwidgets.vlayout()
        leftWidget.setLayout(leftLayout)
        self.splitter.addWidget(leftWidget)


        self.tabFilter = QtWidgets.QTabWidget()
        leftLayout.addWidget(self.tabFilter)

        ##================================
        ## Filter
        grpFilter = xwidgets.GroupGridBox()
        mmm = 5
        grpFilter.setContentsMargins(mmm, mmm, mmm, mmm)
        self.tabFilter.addTab(grpFilter, "Filter")


        # filter radios
        self.buttGroupFilter = QtWidgets.QButtonGroup()
        self.buttGroupFilter.setExclusive(True)

        for ridx, s in enumerate(["Code", "Description", "Code + Description"]):
            rad = QtWidgets.QRadioButton()
            rad.setText(s)
            grpFilter.grid.addWidget(rad, ridx, 0, 1, 2)
            self.buttGroupFilter.addButton(rad, 3 if ridx == 2 else ridx)

        self.buttGroupFilter.button(0).setChecked(True)
        self.buttGroupFilter.buttonClicked.connect(self.on_filter_col)


        # clear button
        self.buttClear = xwidgets.ClearButton(self, callback=self.on_clear_filter)
        grpFilter.grid.addWidget(self.buttClear, 3, 0)

        ## filter text
        self.txtFilter = QtWidgets.QLineEdit()
        self.txtFilter.setMaximumWidth(100)
        grpFilter.grid.addWidget(self.txtFilter, 3, 1)
        self.txtFilter.textChanged.connect(self.on_txt_changed)


        grpFilter.grid.addWidget(QtWidgets.QLabel(), 4, 2)

        #grpFilter.layout.addStretch(3)
        grpFilter.grid.setColumnStretch(0, 0)
        grpFilter.grid.setColumnStretch(1, 10)

        ##================================
        ## Classification Tree
        topLayout = QtWidgets.QVBoxLayout()
        leftLayout.addLayout(topLayout, 0)

        self.treeClass = QtWidgets.QTreeView()
        self.tabFilter.addTab(self.treeClass, "By classification")
        self.treeClass.setModel(AgsData.modelClasses)
        self.treeClass.setRootIsDecorated(False)

        self.treeClass.setExpandsOnDoubleClick(False)

        self.treeClass.setFixedHeight(220)

        self.treeClass.selectionModel().selectionChanged.connect(self.on_class_tree_selected)




        ##== Groups Tree
        self.treeGroups = QtWidgets.QTreeView()
        leftLayout.addWidget(self.treeGroups, 10)
        self.treeGroups.setModel(self.proxy)


        self.treeGroups.setUniformRowHeights(True)
        self.treeGroups.setRootIsDecorated(False)
        self.treeGroups.setAlternatingRowColors(True)

        self.treeGroups.header().setStretchLastSection(True)
        self.treeGroups.setColumnHidden(CG.search, True)
        self.treeGroups.setColumnWidth(CG.code, 120)
        self.treeGroups.setColumnWidth(CG.description, 250)

        self.treeGroups.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.treeGroups.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.treeGroups.setItemsExpandable(False)
        self.treeGroups.setExpandsOnDoubleClick(False)

        self.treeGroups.setSortingEnabled(True)
        self.treeGroups.sortByColumn(CG.code, Qt.AscendingOrder)

        self.treeGroups.selectionModel().selectionChanged.connect(self.on_groups_tree_selected)

        self.agsGroupViewWidget = AGS4GroupViewWidget(self)
        self.splitter.addWidget(self.agsGroupViewWidget)

        self.splitter.setStretchFactor(0, 2)
        self.splitter.setStretchFactor(1, 5)
        Settings.restore_splitter(self.splitter)
        self.splitter.splitterMoved.connect(self.on_splitter_moved)

        #self.statusBar = QtWidgets.QStatusBar()
        #self.mainLayout.addWidget(self.statusBar, 0)


        ##############################################################################
        rightWidget = QtWidgets.QWidget()
        rightLayout = xwidgets.vlayout()
        rightWidget.setLayout(rightLayout)
        self.splitter.addWidget(rightWidget)


        #self.agsHeadingDetailWidget = AGS4HeadingDetailWidget()
        #rightLayout.addWidget(self.agsHeadingDetailWidget)

        #self.init_setup()
        AgsData.sigLoaded.connect(self.on_loaded)

        #self.txtFilter.setText("DETL")

    def on_splitter_moved(self, i, pos):
        Settings.save_splitter(self.splitter)

    def set_focus(self):
        self.txtFilter.setFocus()

    #=========================================
    def on_groups_tree_selected(self, sel=None, desel=None):

        if not self.treeGroups.selectionModel().hasSelection():
             self.agsGroupViewWidget.set_group( None )
             self.sigGroupSelected.emit( None )
             return

        tIdx = self.proxy.mapToSource( sel.indexes()[0] )
        grp_dic = self.proxy.sourceModel().rec_from_midx( tIdx )
        self.agsGroupViewWidget.set_group(grp_dic)
        self.sigGroupSelected.emit( grp_dic )


    def on_filter_col(self, idx):
        self.update_filter()
        self.txtFilter.setFocus()

    def on_txt_changed(self, x):
        self.update_filter()

    def update_filter(self):
        self.treeClass.blockSignals(True)
        self.treeClass.clearSelection()
        self.treeClass.blockSignals(False)

        cidx = self.buttGroupFilter.checkedId()
        self.proxy.setFilterKeyColumn(cidx)

        txt = str(self.txtFilter.text()).strip()
        if "_" in txt:
            grp_code, _ = txt.split("_")
        else:
            grp_code = txt
        self.proxy.setFilterFixedString(grp_code)

        if self.proxy.rowCount() == 1:

            # TODO
            # #self.tree.selectionModel().select(self.proxy.index(0,0))
            pass


    def on_clear_filter(self):
        self.txtFilter.setText("")
        self.txtFilter.setFocus()



    def on_class_tree_selected(self, selected, deselected):
        if not self.treeClass.selectionModel().hasSelection():
            self.txtFilter.setFocus()
            #self.on_group_tree_selected()
            return

        self.proxy.setFilterKeyColumn(CG.cls)

        item = self.treeClass.model().itemFromIndex(selected.indexes()[0])
        if item.text() == "All":
            self.proxy.setFilterFixedString("")
        else:
            self.proxy.setFilterFixedString(item.text())
        self.txtFilter.setFocus()


    def init_load(self):
        pass

    def on_loaded(self):

        ## expand first row
        self.treeClass.setExpanded( self.treeClass.model().item(0,0).index(), True)
        self.treeClass.sortByColumn(0, Qt.AscendingOrder)

        ## set sort orders
        self.treeGroups.sortByColumn(CG.code, Qt.AscendingOrder)
        self.treeGroups.resizeColumnToContents(CG.code)




class AGS4GroupViewWidget( QtWidgets.QWidget ):
    """The GroupView contains the vertically the Group Label at top, headings and notes"""

    sigHeadingSelected = pyqtSignal(object)

    def __init__( self, parent=None, mode=None ):
        QtWidgets.QWidget.__init__( self, parent )

        self.group_code = None

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.topbar = xwidgets.hlayout()
        self.mainLayout.addLayout(self.topbar, 0)

        layInfo = xwidgets.vlayout()
        self.topbar.addLayout(layInfo, 100)

        layInfo2 = xwidgets.hlayout()
        layInfo.addLayout(layInfo2, 0)

        self.icoLabel = xwidgets.IconLabel(self, ico=Ico.AgsGroup)
        self.icoLabel.setStyleSheet("background-color: white; color: #444444;")
        layInfo2.addWidget(self.icoLabel)

        self.lblGroupCode = QtWidgets.QLabel(" ")
        self.lblGroupCode.setStyleSheet("background-color: white; color: %s; font-weight: bold; font-family: monospace; padding: 3px;" % AGS4_COLORS.group)
        self.lblGroupCode.setFixedWidth(50)
        layInfo2.addWidget(self.lblGroupCode)

        lblSpacer = QtWidgets.QLabel(" ")
        lblSpacer.setStyleSheet("background-color: white; color: %s; font-weight: bold; font-family: monospace; padding: 3px;" % AGS4_COLORS.group)
        #self.lblGroupRequired.setFixedWidth(50)
        layInfo2.addWidget(lblSpacer, 100)

        self.lblGroupRequired = QtWidgets.QLabel("Required")
        self.lblGroupRequired.setStyleSheet("background-color: #FFC0CB; color: #444444; font-weight: bold; font-family: monospace; padding: 3px;" )
        self.lblGroupRequired.setVisible(False)
        layInfo2.addWidget(self.lblGroupRequired, 1)

        self.lblDescription = QtWidgets.QLabel(" ")
        self.lblDescription.setStyleSheet("background-color: white; color: #444444; padding: 3px;")
        layInfo.addWidget(self.lblDescription)

        #layInfo.setColumnS

        ##== Parent Child Buttons
        self.buttGrpTB = QtWidgets.QButtonGroup()
        self.buttGrpTB.setExclusive(False)

        self.tbGroupParent = xwidgets.ToolBarGroup(title="Parent")
        self.tbGroupParent.setFixedWidth(70)
        self.topbar.addWidget(self.tbGroupParent, 0)
        self.buttParent = self.tbGroupParent.addButton(text="-", ico=Ico.BulletDown)


        self.tbGroupChildren = xwidgets.ToolBarGroup(title="Children")
        self.topbar.addWidget(self.tbGroupChildren, 0)

        for idx in range(0, 2):
            butt = self.tbGroupChildren.addButton(text="-", ico=Ico.BulletDown)
            butt.setFixedWidth(60)
            self.buttGrpTB.addButton(butt, idx)

        # added after children to make index easy
        self.buttGrpTB.addButton(self.buttParent)
        self.buttGrpTB.buttonClicked.connect(self.on_group_button_clicked)

        self.mainLayout.addSpacing(10)

        ## Headings Table
        self.agsHeadingsTable = AGS4HeadingsTable(self)
        self.mainLayout.addWidget(self.agsHeadingsTable, 10)



        ##== Bottom Splitter
        self.splitBott = QtWidgets.QSplitter()
        self.splitBott.setObjectName("ags_group_view_notes_picklist")
        self.mainLayout.addWidget(self.splitBott)

        ## Notes
        self.agsGroupNotesWidget = AGS4GroupNotesWidget(self)
        self.agsGroupNotesWidget.setFixedHeight(200)
        self.splitBott.addWidget(self.agsGroupNotesWidget)

        ## Abbrs Picklist
        self.agsAbbrevsWidget = AGS4AbbrevsWidget()
        self.splitBott.addWidget(self.agsAbbrevsWidget)


        ## setup splitter
        self.splitBott.setStretchFactor(0, 1)
        self.splitBott.setStretchFactor(1, 1)
        Settings.restore_splitter(self.splitBott)
        self.splitBott.splitterMoved.connect(self.on_splitter_bott_moved)

        self.agsHeadingsTable.sigHeadingSelected.connect(self.on_heading_selection_changed)
        self.agsGroupNotesWidget.sigWordClicked.connect(self.on_word_clicked)

        self.clear_parent_child_buttons()

    def on_group_button_clicked(self, butt):
        group_code = str(butt.text())
        ags4_widgets.show_group_dialog(self, group_code=group_code)

    def on_word_clicked(self, code):
        code = str(code) # WTF!, its a QString not str as sent !
        rec = ags4.DD.words.get(code)
        if rec:

            if rec['type'] == "heading":
                # its a heading, so select it if its in within this group eg SAMP_ID is almost everywhere
                found = self.agsHeadingsTable.select_heading(code)
                if not found:
                    # so its not in this group, so open other group
                    parts = code.split("_")
                    d = AGS4GroupViewDialog(group_code=parts[0], head_code=code)
                    d.exec_()

            if rec['type'] == "group":
                if code != self.group_code:
                    # Dialog only if its not this group
                    d = AGS4GroupViewDialog(group_code=self.group_code)
                    d.exec_()



    def on_splitter_bott_moved(self):
        G.settings.save_splitter(self.splitBott)

    def on_heading_selection_changed(self, head_code):
        self.sigHeadingSelected.emit(head_code)
        self.agsAbbrevsWidget.set_heading(head_code)

    def select_heading(self, head_code):
        self.agsHeadingsTable.select_heading(head_code)
    #
    #
    # def clear(self):
    #     am_i_called___()
    #     self.lblGroupCode.setText("")
    #     self.lblDescription.setText("")
    #     self.agsGroupNotesTable.clear()
    #     self.agsAbbrevsWidget.clear()

    def clear_parent_child_buttons(self):

        for butt in self.buttGrpTB.buttons():
            butt.setIcon(Ico.icon(Ico.BulletBlack))
            butt.setText("")
            butt.setDisabled(True)

    def set_group(self, grp):

        ## load subwidgets, even if grp==None
        self.agsHeadingsTable.set_group(grp)
        self.agsGroupNotesWidget.set_group(grp)
        self.agsAbbrevsWidget.set_heading(None)

        self.clear_parent_child_buttons()
        self.lblGroupRequired.setVisible(False)

        if grp == None:
            self.group_code = None
            self.lblGroupCode.setText("")
            self.lblDescription.setText("")
            return

        self.group_code = grp['group_code']
        self.lblGroupCode.setText(grp['group_code'])
        self.lblDescription.setText(grp['group_description'])


        self.lblGroupRequired.setVisible( grp['group_required'] if grp['group_required'] else False )


        if grp['parent']:
            self.buttParent.setIcon(Ico.icon(Ico.AgsGroup))
            self.buttParent.setText(grp['parent'])
            self.buttParent.setDisabled(False)


        if grp['children']:
            for idx, kid in enumerate(grp['children']):
                butt = self. buttGrpTB.button(idx)
                butt.setIcon(Ico.icon(Ico.AgsGroup))
                butt.setText(kid)
                butt.setDisabled(False)

class AGS4GroupViewDialog(QtWidgets.QDialog):


    def __init__(self, parent=None, group_code=None, head_code=None):
        QtWidgets.QDialog.__init__(self, parent)

        self.setWindowTitle(group_code)
        self.setWindowIcon(Ico.icon(Ico.Ags4))

        self.setMinimumWidth(1100)
        self.setMinimumHeight(500)


        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setSpacing(0)
        margarine = 0
        self.mainLayout.setContentsMargins(margarine, margarine, margarine, margarine)
        self.setLayout(self.mainLayout)



        self.groupViewWidget = AGS4GroupViewWidget(self)
        self.mainLayout.addWidget(self.groupViewWidget)

        grp = ags4.DD.group(group_code)
        self.groupViewWidget.set_group(grp)
        if head_code:
            self.groupViewWidget.select_heading(head_code)



class AGS4HeadingsTable( QtWidgets.QWidget ):

    sigHeadingSelected = pyqtSignal(object)

    def __init__( self, parent ):
        QtWidgets.QWidget.__init__( self, parent )

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        ##===============================================================
        self.tree = QtWidgets.QTreeView(self)
        self.mainLayout.addWidget(self.tree)
        self.tree.setUniformRowHeights(True)
        self.tree.setRootIsDecorated(False)
        self.tree.setAlternatingRowColors(True)

        self.model = HeadingsModel(self)
        self.tree.setModel(self.model)

        CH = HeadingsModel.CH
        self.tree.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Fixed)
        self.tree.setColumnWidth(CH.action, 5)

        self.tree.setColumnWidth(CH.head_code, 100)
        self.tree.setColumnWidth(CH.description, 250)
        self.tree.setColumnWidth(CH.unit, 50)
        self.tree.setColumnWidth(CH.key, 30)
        self.tree.setColumnWidth(CH.required, 30)
        self.tree.setColumnWidth(CH.data_type, 60)
        self.tree.setColumnWidth(CH.sort_order, 20)
        self.tree.header().setStretchLastSection(True)

        self.tree.setSortingEnabled(False)

        self.tree.setContextMenuPolicy( Qt.CustomContextMenu )
        self.tree.customContextMenuRequested.connect(self.on_tree_context_menu )
        self.tree.selectionModel().selectionChanged.connect(self.on_tree_selected)
        self.tree.doubleClicked.connect(self.on_tree_double_clicked)

        self.popMenu = QtWidgets.QMenu()
        self.actOpenGroup = self.popMenu.addAction(Ico.icon(Ico.AgsGroup), "CODEEEEE", self.on_act_open_group)

    def on_tree_context_menu(self, qPoint):
        idx = self.tree.indexAt(qPoint)

        rec = self.model.rec_from_midx(idx)
        gc = rec['head_code'].split("_")[0]
        self.actOpenGroup.setDisabled(gc == self.model.grpDD['group_code'])
        self.actOpenGroup.setText("Open: %s" % gc)
        self.popMenu.exec_(self.tree.mapToGlobal(qPoint))

    def on_act_open_group(self):
        selidx = self.tree.selectionModel().selectedIndexes()
        rec = self.model.rec_from_midx(selidx[0])
        hc = rec.get("head_code")
        gc = hc.split("_")[0]
        d = AGS4GroupViewDialog(self, group_code=gc, head_code=hc)
        d.exec_()


    def set_group(self, grp):
        self.model.set_group(grp)

    def on_tree_double_clicked(self):
        pass

    def on_tree_selected(self, sel, desel):

        if not self.tree.selectionModel().hasSelection():
             self.sigHeadingSelected.emit( None )
             return

        rec = self.model.rec_from_midx( sel.indexes()[0] )
        self.sigHeadingSelected.emit(rec)


    def select_heading(self, head_code):

        midx = self.model.get_heading_index(head_code)
        if midx != None:
            self.tree.selectionModel().setCurrentIndex(midx,
                                                   QtWidgets.QItemSelectionModel.SelectCurrent|QtWidgets.QItemSelectionModel.Rows)


class AGS4GroupNotesWidget( QtWidgets.QWidget ):

    sigWordClicked = pyqtSignal(str)

    def __init__( self, parent=None ):
        QtWidgets.QWidget.__init__( self, parent )


        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        ##==============================
        scrollArea = QtWidgets.QScrollArea()
        scrollArea.setWidgetResizable(True)
        self.mainLayout.addWidget(scrollArea, 100)

        self.scrollWidget = QtWidgets.QWidget()
        scrollArea.setWidget(self.scrollWidget)

        self.scrollLayout = QtWidgets.QVBoxLayout()
        self.scrollLayout.setContentsMargins(0, 0, 0, 0)
        self.scrollLayout.setSpacing(0)
        self.scrollWidget.setLayout(self.scrollLayout)



    def clear(self):
        """Removes all entries"""

        ## pain.. all this shite just to nuke a list
        self.setUpdatesEnabled(False)
        while self.scrollLayout.count() > 0:
            vari = self.scrollLayout.itemAt(0)
            w = vari.widget()
            if w:
                self.scrollLayout.removeWidget( w )
                w.setParent(None)
                w = None
            else:
                self.scrollLayout.removeItem( vari )

        self.setUpdatesEnabled(True)
        self.update()

    def set_group(self, grp):

        self.clear()
        if grp == None:
            return

        notes = grp.get("notes")
        if notes == None:
            return

        self.setUpdatesEnabled(False)
        lookup = ags4.DD.words

        for note in notes:

            w = widget = QtWidgets.QLabel()

            words = note.split(" ")
            res = []
            for word in words:
                if word in lookup:
                    res.append("<a href='#%s-%s'>%s</a>" % (lookup[word]['type'], word, word))
                else:
                    res.append(word)

            widget.setText(" ".join(res))
            widget.setTextFormat(QtCore.Qt.RichText)
            widget.setWordWrap(True)
            widget.setMargin(0)
            sty = "background-color: #EEF1F8; padding: 2px; margin:0; border-bottom:1px solid #dddddd;"
            widget.setStyleSheet(sty)
            widget.setAlignment(QtCore.Qt.AlignTop)

            self.scrollLayout.addWidget(w, 0)

            widget.linkActivated.connect(self.on_link_activated)


        self.scrollLayout.addStretch(10)
        self.setUpdatesEnabled(True)


    def on_link_activated(self, lnkq):
        lnk = str(lnkq)
        parts = lnk[1:].split("-", 1 )
        self.sigWordClicked.emit( parts[1] )



class AGS4AbbrevsWidget( QtWidgets.QWidget ):
    """Shows pickist and abbrevs etc"""

    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )


        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        self.toolbar = xwidgets.hlayout()
        self.mainLayout.addLayout(self.toolbar, 0)

        self.icoLabel = xwidgets.IconLabel(self, ico=Ico.AgsGroup)
        self.icoLabel.setStyleSheet("background-color: white; color: #444444;")
        self.toolbar.addWidget(self.icoLabel, 0)

        self.lblAbbrCode = QtWidgets.QLabel(" ")
        self.lblAbbrCode.setStyleSheet("background-color: white; color: %s; font-weight: bold; font-family: monospace; padding: 3px;" % AGS4_COLORS.group)
        self.toolbar.addWidget(self.lblAbbrCode, 20)


        ##=== Tree
        self.tree = QtWidgets.QTreeView()
        self.mainLayout.addWidget(self.tree)
        self.tree.setUniformRowHeights(True)
        self.tree.setRootIsDecorated(False)
        self.tree.setAlternatingRowColors(True)
        self.tree.setSortingEnabled(False)

        self.model = AbbrevItemsModel(self)
        self.tree.setModel(self.model)

        CA = AbbrevItemsModel.CA

        self.tree.setColumnWidth(CA.code, 80)
        self.tree.setColumnWidth(CA.list, 50)

        self.tree.setColumnHidden(CA.check, True)

        self.tree.header().setStretchLastSection(True)

        # TODO fix sort to ags
        self.tree.setSortingEnabled(True)

        self.set_heading(None)


    def set_heading(self, heading):
        self.model.set_ags_heading(heading)
        s = ""
        if heading and heading["data_type"] == "PA":
            s = "%s" % (heading['head_description'])
        self.lblAbbrCode.setText(s)



class DecimalEditDelegate(QtWidgets.QItemDelegate):
    """Number editor to n decimal places"""
    def __init__(self, parent, heading):
        QtWidgets.QItemDelegate.__init__(self, parent)

        self.ogtHeading = heading

        ##self.data_type = heading['type']
        #self.data_type = heading['type']
        self.dp = None
        if self.ogtHeading.data_type.endswith("DP"):
            self.dp = int(self.ogtHeading.data_type[:-2])

    def createEditor(self, parent, option, index):

        editor = QtWidgets.QLineEdit(parent)
        if self.ogtHeading.type.endswith("DP"):
            validator = QtWidgets.QDoubleValidator()
            validator.setDecimals(self.dp)
            editor.setValidator(validator)

        return editor

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        curr = index.model().data(index) #.toString()
        editor.setText(curr)
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        no = float(editor.text())
        f = "%01."
        f += "%sf" % self.dp
        txt = f % (no,)
        model.setData(index, txt)

class IDComboDelegate(QtWidgets.QItemDelegate):
    """A combobox for the ID"""
    def __init__(self, parent, heading, options):
        QtWidgets.QItemDelegate.__init__(self, parent)

        self.ogtHeading = heading
        self.options = options

    def createEditor(self, parent, option, index):

        editor = QtWidgets.QComboBox(parent)
        editor.addItem("--unknown--", "")

        # populate combobox from abbreviations
        for v in self.options:
            editor.addItem( "%s" % v, "%s" % v)

        return editor

    def setEditorData(self, editor, index):
        editor.blockSignals(True)
        curr = index.model().data(index).toString()
        idx = editor.findData(curr)
        if idx != -1:
            editor.setCurrentIndex(idx)
        editor.blockSignals(False)

    def setModelData(self, editor, model, index):
        txt = editor.itemData(editor.currentIndex()).toString()
        model.setData(index, txt)

class AGS4UnitsTypesWidget( QtWidgets.QWidget ):
    """The Units and Types tab"""

    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.setObjectName("AGS4UnitTypesWidget")


        self.proxyUnits = QtCore.QSortFilterProxyModel()
        self.proxyUnits.setSourceModel(AgsData.modelUnits)
        self.proxyUnits.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.proxyUnits.setSortCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self.proxyTypes = QtCore.QSortFilterProxyModel()
        self.proxyTypes.setSourceModel(AgsData.modelDataTypes)
        self.proxyTypes.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.proxyTypes.setSortCaseSensitivity(QtCore.Qt.CaseInsensitive)

        ##===============================================
        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.splitter = QtWidgets.QSplitter()
        self.mainLayout.addWidget(self.splitter)

        self.treeUnits = self.make_tree(self.proxyUnits, "Unit", "Description")
        self.splitter.addWidget(self.treeUnits)

        self.treeTypes = self.make_tree(self.proxyTypes, "Type", "Description")
        self.splitter.addWidget(self.treeTypes)


    def make_tree(self, model, tit1, tit2):
        tree = QtWidgets.QTreeView()
        tree.setRootIsDecorated(False)
        tree.setSortingEnabled(True)
        tree.setModel(model)
        return tree
        hi = tree.headerItem()
        hi.setText(0, tit1)
        hi.setText(1, tit2)
        return tree



# -*- coding: utf-8 -*-

import os
from Qt import QtGui, QtCore, QtWidgets, Qt, pyqtSignal


from ogt import FORMATS
from ogt import ogt_doc
from ogt import ogt_cell

import app_globals as G

from img import Ico
import xwidgets

import ogtgui_doc
import ogtgui_groups
import ogtgui_widgets
import ogtgui_errors
import ogtgui_schedules
import ogtgui_samples
import ogtgui_sources
import ogtgui_dialogs
import ogtgui_excel

class OGTProjectWidget( QtWidgets.QWidget ):

    sigChanged = pyqtSignal()

    def __init__( self, parent=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogt_doc.OGTDocument()
        self.ogtDoc.sigChanged = self.sigChanged

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        ##====== Top Bar ===
        lbl = QtWidgets.QLabel()
        lbl.setStyleSheet("background-color: #aaaaaa")
        lbl.setFixedHeight(1)
        self.mainLayout.addWidget(lbl)

        self.topLay = xwidgets.hlayout()
        self.mainLayout.addLayout(self.topLay)

        ## Header Label
        self.headerWidget = ogtgui_widgets.HeaderBarWidget(title="-", bg="#444444", bg2="#999999", fg="#eeeeee")
        self.topLay.addWidget(self.headerWidget, 100)

        lbl = QtWidgets.QLabel()
        lbl.setStyleSheet("background-color: #aaaaaa")
        lbl.setFixedHeight(1)
        self.mainLayout.addWidget(lbl)


        ## Add button
        self.buttActAdd = xwidgets.XToolButton(text="Add..", ico=Ico.Add, popup=True)
        self.topLay.addWidget(self.buttActAdd)
        self.buttActAdd.setMenu(QtWidgets.QMenu())

        #act = QtWidgets.QAction(Ico.icon(Ico.Groups), "Add required groups (PROJ, UNIT, etc)", self.on_add_default_groups)
        self.buttActAdd.menu().addAction(Ico.icon(Ico.Groups), "Add required groups (PROJ, UNIT, etc)", self.on_add_required_groups)

        ## Import button
        self.buttImport = xwidgets.XToolButton(text="Import", ico=Ico.Import, menu=True, popup=True)
        self.topLay.addWidget(self.buttImport)
        self.buttImport.menu()
        self.buttImport.menu().addAction(Ico(Ico.Excel), "Import excel", self.on_import_excel)

        ## Validate button
        self.buttValidate = xwidgets.XToolButton(text="Re-Validate", ico=Ico.Refresh, callback=self.on_validate)
        self.topLay.addWidget(self.buttValidate)

        ## Export button
        self.buttExport = xwidgets.XToolButton(text="Save & Export", ico=Ico.Export, callback=self.on_save_export)
        self.topLay.addWidget(self.buttExport)
        #for a in FORMATS:
        #    self.buttExport.menu().addAction("%s - TODO" % a)

        ## Reload button
        #self.buttReload = xwidgets.XToolButton(text="Reload", ico=Ico.Refresh, callback=self.on_reload)
        #self.topLay.addWidget(self.buttReload)



        self.mainLayout.addSpacing(5)

        ##========= Content ===============

        ## tabar + Stack
        tbarLay = xwidgets.hlayout(margin=0)
        self.mainLayout.addLayout(tbarLay)

        self.tabBar = QtWidgets.QTabBar()
        f = self.tabBar.font()
        f.setBold(True)
        self.tabBar.setFont(f)
        tbarLay.addWidget(self.tabBar)
        tbarLay.addStretch(5)

        self.stackWidget = QtWidgets.QStackedWidget()
        self.mainLayout.addWidget(self.stackWidget)

        ## Summary Tab
        self.tabBar.addTab(Ico.icon(Ico.Summary), "Summary")
        self.ogtProjSummaryWidget = OGTProjectSummaryWidget(ogtDoc=self.ogtDoc)
        self.stackWidget.addWidget(self.ogtProjSummaryWidget)
        self.ogtProjSummaryWidget.sigGoto.connect(self.on_goto)
        #self.ogtProjSummaryWidget.sigGotoSource.connect(self.on_goto_source)

        ## Groups Tab
        self.tabBar.addTab(Ico.icon(Ico.Groups), "Groups")
        self.ogtDocWidget = ogtgui_doc.OGTDocumentWidget(ogtDoc=self.ogtDoc)
        self.stackWidget.addWidget(self.ogtDocWidget)

        #self.ogtDoc.sigChanged.connect(self.on_changed)
        #self.ogtDocWidget.sigChanged.connect(self.on_changed)


        ## Samples Tab
        self.tabBar.addTab(Ico.icon(Ico.Samples), "Samples")
        self.ogtSamplesWidget = ogtgui_samples.OGTSamplesWidget(ogtDoc=self.ogtDoc)
        self.stackWidget.addWidget(self.ogtSamplesWidget)


        ## Schedule Tab
        self.tabBar.addTab(Ico.icon(Ico.Schedule), "Schedule")
        self.ogtScheduleWidget = ogtgui_schedules.OGTScheduleWidget(ogtDoc=self.ogtDoc)
        xidx = self.stackWidget.addWidget(self.ogtScheduleWidget)



        ## Source tab
        self.tabBar.addTab(Ico.icon(Ico.Source), "Source")
        self.ogtSourceViewWidget = ogtgui_sources.OGTSourceViewWidget(ogtDoc=self.ogtDoc)
        self.stackWidget.addWidget(self.ogtSourceViewWidget)

        """
        if False:
            self.tabBar.addTab(Ico.icon(Ico.Map), "Map")
            self.mapOverviewWidget = map_widgets.MapOverviewWidget()
            self.stackWidget.addWidget(self.mapOverviewWidget)
        """

        self.tabBar.currentChanged.connect(self.on_tab_changed)

        self.sigChanged.connect(self.on_changed)

        if G.args.dev:
            self.tabBar.setCurrentIndex(1)
            pass

    def init_load(self):
        pass

    def on_tab_changed(self, idx):
        self.stackWidget.setCurrentIndex(idx)
        self.stackWidget.widget(self.stackWidget.currentIndex()).do_update()


    def on_changed(self):
        #print("on_changed", self)
        #print(self.sender())
        self.on_validate()

    def on_butt_refresh_sched(self):
        self.ogtDoc.validate()
        self.ogtScheduleWidget.do_update()


    def on_validate(self):
        self.ogtDoc.validate()
        for widget in [self.ogtDocWidget]:
            widget.do_update()


    def add_ags4_file(self, file_path):

        #self.file_path = None
        """
        self.doc, err = ogt_doc.OGTDocument()
        err = self.doc.load_from_ags4_file(file_path)

        """
        #err = ogt_doc.create_doc_from_ags4_file(file_path)

        err = self.ogtDoc.add_ags4_file(file_path)

        # TODO
        self.ogtDoc.validate()
        self.do_update()

    def on_add_stuff(self):

        dial = ogtgui_dialogs.FileSelectDialog(self)
        dial.exec_()


    def on_add_test_ags(self):
        fn = "/home/geo2lab/z_drive/jobs/40101-40200/40153/AGS/40153 AGS.ags"
        self.do_update()

    def do_update(self):
        #proj = None #self.ogtDoc.proj_dict()
        proj = self.ogtDoc.proj_dict()

        if proj:
            self.headerWidget.setTitle(proj.get('PROJ_NAME', "PROJ_NAME is missing"))


        self.ogtDocWidget.do_update()
        self.ogtProjSummaryWidget.do_update()
        self.ogtSamplesWidget.do_update()
        self.ogtScheduleWidget.do_update()
        self.ogtSourceViewWidget.do_update()


        ## HACK
        #QtCore.QTimer.singleShot(4000, self.do_map_after)

    def do_map_after(self):
        self.mapOverviewWidget.load_document(self.ogtDoc)

    def on_goto(self, code):

        self.ogtDocWidget.select_group(code)
        idx = self.stackWidget.indexOf(self.ogtDocWidget)
        self.tabBar.setCurrentIndex(idx)



    def on_goto_source(self, lidx, cidx):

        self.ogtSourceViewWidget.select_cell(lidx, cidx)
        idx = self.stackWidget.indexOf(self.ogtSourceViewWidget)
        self.tabBar.setCurrentIndex(idx)


    def on_add_required_groups(self):

        self.ogtDoc.add_required_groups()


    def on_import_excel(self):
        #file_path = os.getcwd()
        file_path = "/home/ogt"

        dial = ogtgui_dialogs.FileSelectDialog(self, root=file_path)
        if dial.exec_():
            file_path = dial.file_path

            w = ogtgui_excel.ExcelBrowserDialog(self, file_path=file_path)
            w.exec_()

    def on_save_export(self):

        dial = ogtgui_dialogs.SaveExportDialog(self, self.ogtDoc)
        dial.exec_()

class CP:
    node = 0
    group_code = 1
    group_description = 2



class OGTProjectsModel(QtCore.QAbstractItemModel):

    class C:
        node = 0
        group_code = 1
        group_description = 2

    def __init__( self, parent=None):
        QtCore.QAbstractItemModel.__init__( self, parent )

        self.ogtDoc = None

    def load_document(self, ogtDoc):
        self.ogtDoc = ogtDoc
        self.modelReset.emit()

    def columnCount(self, pidx):
        return 3

    def headerData(self, p_int, orientation, role=None):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            lst = ["Rows", "Group", "Description"]
            return QtCore.QVariant(lst[p_int])
        return QtCore.QVariant()

    def rowCount(self, pidx):
        if self.ogtDoc == None:
            return 0
        if pidx.row() == -1:
            return self.ogtDoc.groups_count()
        return 0 #self.ogtDoc.groups_count()

    def index(self, row, col, pidx):
        return self.createIndex(row, col, None)
        #return QtCore.QModelIndex()

    def parent(self, index):

        if not index.isValid():
            return QtCore.QModelIndex()
        node = index.internalPointer()
        if node != None and node.parent() is None:
            ss
            return QtCore.QModelIndex()
        else:
            return self.createIndex(0, 0, None)
            return self.createIndex(node.parent.row, 0, node.parent)


        if False: #child.isValid():

            ip =  child.internalPointer()

            if ip:
                return self.createIndex(ip.parent().row(), 0, ip.parent())
            #return QtCore.QModelIndex()
        #return self.createIndex(child.row(), child.column(), None)
        return QtCore.QModelIndex()

    def data(self, midx, role):

        if role == Qt.DecorationRole:
            return QtCore.QVariant()

        if role == Qt.TextAlignmentRole:
            return QtCore.QVariant(int(Qt.AlignTop | Qt.AlignLeft))

        if role != Qt.DisplayRole:
            return QtCore.QVariant()

        #node = self.nodeFromIndex(midx)

        if midx.column() == 0:
            return QtCore.QVariant(self.ogtDoc.group_by_index(midx.row()))

        elif midx.column() == 1:
            return QtCore.QVariant(self.ogtDoc.group_by_index(midx.row()))

        elif midx.column() == 2:
            return QtCore.QVariant(self.ogtDoc.group_by_index(midx.row()))

        return QtCore.QVariant()




class OGTProjectSummaryWidget( QtWidgets.QWidget ):

    sigGoto = pyqtSignal(object)
    #sigGotoSource = pyqtSignal(int, int)

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogtDoc

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.headerWidget = ogtgui_widgets.HeaderBarWidget(title="Project Summary", bg="#56B8F9")
        self.mainLayout.addWidget(self.headerWidget, 0)

        self.splitter = QtWidgets.QSplitter()
        self.mainLayout.addWidget(self.splitter, 100)


        self.treeSummary = QtWidgets.QTreeWidget()
        hi = self.treeSummary.headerItem()
        hi.setText(0, "Summary")
        self.splitter.addWidget(self.treeSummary)



        self.errorsWidget = ogtgui_errors.ErrorsWidget(mode=ogtgui_errors.VIEW_ERR_MODE.document, ogtDoc=self.ogtDoc)
        self.splitter.addWidget(self.errorsWidget)
        #self.errorsWidget.sigGotoSource.connect(self.on_goto_source)
        self.errorsWidget.sigGotoError.connect(self.on_goto_error)

        self.groupsListWidget = ogtgui_groups.GroupsListWidget(ogtDoc=self.ogtDoc)
        self.splitter.addWidget(self.groupsListWidget)

    def clear(self):

        self.errorsWidget.clear()
        self.tree.clear()

    def do_update(self):

        self.errorsWidget.do_update()
        self.groupsListWidget.do_update()

        self.treeSummary.clear()
        item = xwidgets.XTreeWidgetItem()
        item.setText(0, "TODO")
        self.treeSummary.addTopLevelItem(item)



    def on_tree_double_clicked(self, item, cidx):

        item = self.tree.currentItem()
        if item == None:
            return
        self.sigGoto.emit(item.text(CP.group_code))

    def DEADon_goto_source(self, lidx, cidx):
        self.sigGotoSource.emit(lidx, cidx)

    def on_goto_error(self, ogtErr):
        return
        self.sigGotoError.emit(ogtErr)

#
# class XStackedWidget( QtWidgets.QWidget ):
#
#     """Psuedo Stack as it containes a header/label and widget"""
#     def __init__( self, parent=None):
#         QtWidgets.QWidget.__init__( self, parent )
#
#         self.mainLayout = QtWidgets.QVBoxLayout()
#         self.mainLayout.setContentsMargins(0,0,0,0)
#         self.mainLayout.setSpacing(4)
#         self.setLayout(self.mainLayout)
#
#         self.headerStack = QtWidgets.QStackedWidget()
#         self.mainLayout.addWidget(self.headerStack, 0)
#
#         self.contentStack = QtWidgets.QStackedWidget()
#         self.mainLayout.addWidget(self.contentStack, 100)
#
#     def addWidget(self, widget, header_text, bg="#dddddd"):
#
#         tbar = QtWidgets.QWidget()
#         tlay = QtWidgets.QHBoxLayout()
#         tlay.setContentsMargins(0,0,0,0)
#         tlay.setSpacing(0)
#         tbar.setLayout(tlay)
#
#         lbl = QtWidgets.QLabel()
#         lbl.setText(header_text)
#         sty = " color: #666666; font-size: 14pt; padding: 2px 5px;"
#         sty += "background-color: qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, "
#         sty += "stop: 0 #efefef "
#         sty += ", stop: 0.3 #efefef "
#         sty += "stop: 1 %s" % bg
#         sty += ");"
#         lbl.setStyleSheet(sty)
#         tlay.addWidget(lbl, 10)
#
#         self.headerStack.addWidget(tbar)
#
#         nidx = self.contentStack.addWidget(widget)
#
#         return nidx
#
#     def setCurrentIndex(self, idx):
#         self.headerStack.setCurrentIndex(idx)
#         self.contentStack.setCurrentIndex(idx)
#
#     def addHeaderWidget(self, idx, widget):
#
#         wid = self.headerStack.widget(idx)
#         wid.layout().addWidget(widget, 0)
class OGTProjectOverviewWidget( QtWidgets.QWidget ):

    sigGoto = pyqtSignal(object)
    #sigGotoSource = pyqtSignal(int, int)

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogtDoc

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.treeWidget = QtWidgets.QTreeWidget()

        self.mainLayout.addWidget(self.treeWidget)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("ogtstatic/version.txt", "r") as fh:
    version = fh.read()

setuptools.setup(
    name="open-geotechnical", # Replace with your own username
    version=version,

    author="Pedro Morgan",
    author_email="ogt@daffodil.uk.com",

    description="Open Geotechnical",
    long_description=long_description,

    url="https://open-geotechnical.gitlab.io/ogt_py/",
    project_urls={
        "Bug Tracker": "https://gitlab.com/open-geotechnical/ogt_py/issues",
        "Documentation": "https://open-geotechnical.gitlab.io/ogt_py/",
        "Source Code": "https://gitlab.com/open-geotechnical/ogt_py",
    },

    packages=setuptools.find_packages(exclude="docs"),
    include_package_data=True,
    package_data={
        'ogtstatic': ['*'],
    },

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],

    python_requires='>=3.4',
    install_requires=[
       'mistune',
       'openpyxl',
        "pyyaml",
        "geojson",
        "flask"
    ],

    scripts=[
        'ogt-cli.py',
        'ogt-desktop.py',
        'ogt-server.sh'
    ],
    data_files=[
        ('share/icons/32x32/apps', ['ogtstatic/icons/open-geotechnical.png']),
        ('share/applications', ['etc/open-geotechnical.desktop']),
    ],
)



# coding: utf-8

from ogt import ags4
from ogt import CELL_COLORS


class OGTError:
    """"""
    WARN = 0
    ERR = 1
    OK = 5

    def __init__(self, message=None, cell=None, warn=False, rule=None):
        """Contains an error or warning

        .. code-block:: python

            cell = OGTCell(raw=" Samp ")

            if test_some_warning(cell.value):
                cell.add_error( OGTError(cell, "some not valid", warn=True)

            if test_some_error(cell.value):
                cell.add_error( OGTError(cell, "WTF !!")

            print( cell.errors )


        :param cell: reference to parent cell
        :type cell: OGTCell
        :param message: message string
        :type message: str
        :param warn: True if its not an error
        :type warn: bool
        :param rule: optional rule message
        :type rule: str or int
        """
        assert message is not None

        self.cell = cell
        """The parent cell of this error

        :type: OGTCell
        """

        self.err_type = OGTError.WARN if warn else OGTError.ERR
        """True to flag as error(default), False is a warning"""

        self.message = message
        """The error message

        :type: str
        """

        self.rule = None if rule is None else str(rule)
        """The ags4 rule of error"""

    def __repr__(self):
        return "<OGTError:%s: %s>" % ("ERR" if self.err_type else "WARN", self.message)


    @property
    def group_code(self):
        """"""
        return self.cell.parentHeading.parentGroup.group_code

    @property
    def head_code(self):
        """"""
        return self.cell.parentHeading.head_code

    @property
    def background(self):
        """Background color of error/warning"""
        return CELL_COLORS.err_bg if self.err_type == OGTError.ERR else CELL_COLORS.warn_bg


def to_upper(cell):
    """"""
    pre_up = str(cell.value)
    cell.value = cell.value.upper()
    if pre_up != cell.value:
        cell.add_error(OGTError(cell, "Not upper case `%s`" % pre_up, warn=True))


def ags4_descriptor(descriptor):
    """Check its case and whether its a data descriptor

    :return: Cleaned descriptor, valid_descroiptor and list of errors

    """

    # check upper
    up_des, errs = validate_upper(descriptor)

    # check only a2z
    cerrs = validate_a2z(up_des)
    if cerrs:
        errs.extend(cerrs)

    # check it a description
    if up_des in ags4.DD.descriptors():
        return up_des, True, errs

    errs.append(OGTError("Invalid descriptor `%s` " % up_des),
                warn=True, rule=3)
    return up_des, False, errs


def heading_code(group_code, cell):
    """"""
    grp = ags4.DD.group(group_code)
    if cell.value in grp['headings_index']:
        return

    mess = "Invalid heading `%s` in group `%s` " % (cell.value, group_code)
    cell.add_error(OGTError(cell, mess))


def data_type(cell):
    """"""
    types_dic = ags4.DD.data_types_dict
    if cell.value not in types_dic:
        mess = "Invalid type `%s` " % (cell.value)
        cell.add_error(OGTError(cell, mess))


def validate_clean_str(raw_code, upper=True):
    """
    :param raw_code: str to clean
    :type raw_code: str
    :param upper: Make string uppercase
    :type upper: bool
    :rtype: (str, [OGTError])
    :return: Stripped and optinal upper case
    """

    assert isinstance(raw_code, str)

    err_list = []

    # Check for whitespace
    code = str(raw_code).strip()
    if code != raw_code:
        if code == raw_code.lstrip():
            e = OGTError("Leading white space `%s`" % raw_code, warn=True)
            err_list.append(e)
        elif code == raw_code.rstrip():
            e = OGTError("Trailing white space `%s`" % raw_code, warn=True)
            err_list.append(e)
        else:
            e = OGTError("White space`%s`" % raw_code, warn=True)
            err_list.append(e)

    if not upper:
        return code, err_list


    ucode, xerrs = validate_upper(code)
    if len(xerrs) > 0:
        err_list.extend(xerrs)

    return ucode, err_list


def validate_upper(raw_str):
    """Validate uppercase

    :param raw_str:
    :type raw_str: str
    :rtype: (str, [OGTError])
    :return: Upper case string
    """
    err_list = []
    ustr = raw_str.upper()
    if ustr != raw_str:
        e = OGTError("Contains lower chars `%s`" % raw_str, warn=True)
        err_list.append(e)
    return ustr, err_list


A2Z = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
NUMBERS = "0123456789"
CHARS = A2Z + NUMBERS


def validate_a2z(astr):
    """Ensure string contains only A-Z uppercase

    :return: A list of errors
    """
    errs = []
    for idx, char in enumerate(astr):
        if char not in A2Z:
            errs.append(OGTError("Invalid char `%s` at position %s `%s`" % (char, idx + 1, astr), rule=19))
    return errs


def validate_group_code(group_code):
    """ :ref:`ags4_rule_19`
    :param group_code: group code to validate
    :rtype: (str, [OGTError])
    :return: Tuple with cleaned string and errors
    """

    group_code, errs = validate_clean_str(group_code, upper=True)

    # first check length for empty
    lenny = len(group_code)
    if lenny == 0:
        # no group code so get outta here
        errs.append(OGTError("Invalid GROUP, needs one char at least `%s`" % group_code,
                            rule=19))
        return group_code, errs

    # check over length
    if lenny > 4:
        errs.append(OGTError("Invalid GROUP, longer then four chars `%s`" % group_code,
                             rule=19))
        return group_code, errs

    # check characters are valid
    errs = []
    for idx, char in enumerate(group_code):
        if char in A2Z or char in NUMBERS:
            pass  # ok
        else:
            errs.append(OGTError("Invalid char in  GROUP position %s `%s`" % (idx+1, group_code),
                                 rule=19))

    return group_code, errs


def validate_heading_str(head_code):
    """Checks the heading is valid
    - cleaned code
    - whether fatal
    - errors list
    """

    head_code, errs = validate_upper(head_code)

    if "_" not in head_code:
        errs.append(OGTError("Invalid HEADING requires a _ `%s` not found" % head_code))
        # cannot continue ??
        return head_code, True, errs

    # split the heading into group + remainder
    group_code, head_part = head_code.split("_")

    # validate the group part
    group_code, errs = validate_group_code(group_code)
    if len(errs) > 0:
        # assume we cannot continue
        return group_code, True, errs

    if len(head_code) > 9:
        errs.append(OGTError("Invalid HEADING > 9 chars `%s`" % head_code,
                             rule="19a"))

    if len(head_code) == 0:
        errs.append(OGTError("Invalid HEADING needs at least one char `%s`" % head_code,
                             rule="19a"))

    return head_code, False, errs


def validate_heading_ags(head_code, group_code_in):
    """Check the heading is in ags data dict. This is done by
         - First check if the heading is in the group (eg SPEC_*)
         - Then split the HEAD_ING and check the group and the heading in theat group

    """
    # first get the group that this heading is in
    # check group exists
    grpdic = ags4.DD.group(group_code_in)

    if grpdic is None:
        return OGTError("Invalid GROUP `%s` for HEADING `%s`, not in ags data dict" % (group_code_in, head_code),
                        rule="9")

    heads = grpdic.get("headings")
    for h in heads:
        if head_code == h['head_code']:
            # Yipee, the head_code is in the origin group
            return None

    # split the head code amd check source group + heading i that group

    sgroup_code, _ = head_code.split("_")

    # check group exists
    grpdic = ags4.DD.group(sgroup_code)

    if grpdic is None:
        return OGTError("Invalid GROUP part of HEADING not in ags data dict `%s`" % head_code,
                        rule=9)

    heads = grpdic.get("headings")
    for h in heads:
        if head_code == h['head_code']:
            return None
    return OGTError("HEADING `%s` not found in GROUP `%s`" % (head_code, sgroup_code),
                    warn=True, rule=9)


def validate_type_ags(typ):

    types = ags4.DD.data_types_dict
    if typ in types:
        return None

    return OGTError("TYPE `%s` not ing AGS4 " % (typ), rule="TODO")


def validate_headings_sort(group_code, heading_codes):

    # get ags headings list for group
    ags_headings = ags4.DD.headings_index(group_code)
    if ags_headings is None:
        return

    plst = []
    for ags_head in ags_headings:
        if ags_head in heading_codes:
            plst.append(ags_head)
        # else:
        #    plst.append(None)
    if plst == heading_codes:
        return None

    clst = []
    cidxs = []

    for idx, c in enumerate(plst):
        if c is None:
            continue
        if heading_codes[idx] != c:
            clst.append(c)
            cidxs.append(idx + 1)
    if len(clst) == 0:
        return None
    errs = []
    for idx, cidx in enumerate(cidxs):
        mess = "heads order incorrect, should be `%s` cols %s " % (",".join(clst), ",".join([str(c) for c in cidxs]))
        errs.append(OGTError(message=mess, rule=9))
    return errs



# -*- coding: utf-8 -*-



import collections
from Qt import Qt, QtCore, QtGui, QtWidgets, pyqtSignal

import ogtgui_widgets
import xwidgets
from img import Ico

HOVER_COLOR = "#EBEBEB"

class OGTSamplesTableModel(QtCore.QAbstractTableModel):
    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractTableModel.__init__(self)

        self.ogtDoc = ogtDoc
        self.loca_id = None

        self.cidx = None
        self.ridx = None

    def set_current(self, ridx=None, cidx=None):
        self.ridx = ridx
        self.cidx = cidx

    def row_index_from_samp_id(self, samp_id):
        grp = self.ogtDoc.group_from_code("SAMP")
        # cidx = grp.index_of_heading("SAMP_ID")
        for ridx, row in enumerate(grp.data):
            cell = row.get("SAMP_ID")
            #print cell.value, samp_id, cell.value == samp_id
            if cell.value == samp_id:
                return ridx
        return None

    def get_index_from_cell(self, scell):
        grp = self.ogtDoc.group_from_code("SAMP")
        for ridx in range(0, grp.data_rows_count):
            for cidx in range(0, grp.headings_count):
                cell = grp.data_cell(ridx, cidx)
                if cell == scell:
                    return self.index(ridx, cidx)
        return None

    def columnCount(self, pidx=None):
        grp = self.ogtDoc.group_from_code("SAMP")
        if grp:
            return grp.headings_count
        return 0

    def rowCount(self, pidx=None):
        grp = self.ogtDoc.group_from_code("SAMP")
        if grp:
            return grp.data_rows_count
        return 0

    def data_row(self, row):
        return self.ogtDoc.group_from_code("SAMP").data[row]

    def data(self, midx, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = midx.row()
        col = midx.column()

        if role == Qt.DisplayRole:
            cell = self.ogtDoc.group_from_code("SAMP").data_cell(row, col)
            # print cell
            return cell.value

        if role == Qt.BackgroundRole:
            bg = "white"
            if midx.row() == self.ridx:
                bg = HOVER_COLOR
            return QtGui.QColor(bg)


        return None

    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            return self.ogtDoc.group_from_code("SAMP").heading_from_index(idx).head_code

        if role == Qt.DisplayRole and orient == Qt.Vertical:
            return idx + 1

        if role == Qt.TextAlignmentRole and orient == Qt.Horizontal:
            return Qt.AlignRight if idx == 0 else Qt.AlignLeft

        return None



class OGTScheduleTestsModel(QtCore.QAbstractTableModel):

    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractTableModel.__init__(self)

        self.xheads = None

        self.ridx = None
        self.cidx = None

    def set_current(self, ridx=None, cidx=None):
        self.ridx = ridx
        self.cidx = cidx

    def rowCount(self, pidx=None):
        if self.xheads == None:
            return 0
        return 1

    def columnCount(self, pidx=None):
        if self.xheads == None:
            return 0
        return len(self.xheads)

    def to_no(self, v ):
        try:
            return float(v)
        except:
            return None

    def data(self, midx, role=Qt.DisplayRole):

        if role == Qt.DisplayRole:
            # only one row
            return self.xheads[midx.column()]

        if role == Qt.TextAlignmentRole:
            return Qt.AlignTop|Qt.AlignCenter

        if role == Qt.BackgroundRole:
            bg = "white"
            if midx.column() == self.cidx:
                bg = HOVER_COLOR
            return QtGui.QColor(bg)


    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Vertical:
            return " "

class OGTScheduleTableModel(QtCore.QAbstractTableModel):

    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractTableModel.__init__(self)

        self.xcells = None
        self.xheads = None
        self.row_counts = None
        self.col_counts = None

        self.hi_row = None
        self.hi_col = None

    def set_current(self, ridx, cidx):
        self.curr = [ridx, cidx]
        self.layoutChanged.emit()

    def rowCount(self, pidx=None):
        if self.xheads == None:
            return 0
        return len(self.xcells)

    def columnCount(self, pidx=None):
        if self.xheads == None:
            return 0
        return len(self.xheads)

    def to_no(self, v ):
        try:
            return float(v)
        except:
            return None


    def data(self, midx, role=Qt.DisplayRole):

        if False and role == Qt.DisplayRole or role == Qt.EditRole:
            cell = self.xcells[midx.row()][midx.column()]
            if cell:
                return "YES"

        if False and role == Qt.CheckStateRole:
            cell = self.xcells[midx.row()][midx.column()]
            if cell:
                return Qt.Checked
            return Qt.Unchecked

        if  False and role == Qt.BackgroundRole:
            #cell = self.xcells[midx.row()][midx.column()]
            #if cell:
            #    return QtWidgets.QColor("orange")

            if self.hi_row != None:
                if self.hi_row == midx.row():
                    return QtGui.QColor("#eeeeee")

    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            v = self.col_counts.get(idx)
            if v != None:
                return v
            return ""

        if role == Qt.DisplayRole and orient == Qt.Vertical:
            v = self.row_counts.get(idx)
            if v != None:
                return v
            return ""

        if role == Qt.TextAlignmentRole:
            return Qt.AlignCenter

        return None

class OGTScheduleWidget( QtWidgets.QWidget ):
    """The SourceViewWidget info which in row 0 """

    TOP_HEIGHT = 100

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogtDoc

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.headerBar = ogtgui_widgets.HeaderBarWidget(title="Schedule", bg="#B3C256")
        self.mainLayout.addWidget(self.headerBar, 0)

        self.buttRefreshSched = QtWidgets.QToolButton()
        self.buttRefreshSched.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.buttRefreshSched.setText("Reload")
        self.buttRefreshSched.setIcon(Ico.icon(Ico.Refresh))
        self.headerBar.addWidget(self.buttRefreshSched)
        self.buttRefreshSched.clicked.connect(self.on_butt_refresh)

        self.splitter = QtWidgets.QSplitter()
        self.mainLayout.addWidget(self.splitter, 200)
        self.splitter.setOrientation(Qt.Horizontal)

        ### left -------------------------

        self.leftWidget = QtWidgets.QWidget()
        self.splitter.addWidget(self.leftWidget )
        self.leftLay = xwidgets.vlayout()
        self.leftWidget.setLayout(self.leftLay)

        self.lblCorner = QtWidgets.QLabel()
        self.lblCorner.setText("---")
        self.lblCorner.setFixedHeight(self.TOP_HEIGHT)
        self.leftLay.addWidget(self.lblCorner)

        ## == Samples
        self.tableSamples = XTableHoverView()
        self.leftLay.addWidget(self.tableSamples, 100)

        self.modelSamples = OGTSamplesTableModel(ogtDoc=self.ogtDoc)
        self.tableSamples.setModel(self.modelSamples)

        self.tableSamples.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableSamples.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableSamples.selectionModel().selectionChanged.connect(self.on_tree_samples_selection)

        #self.tableSamples.verticalScrollBar().hide()
        self.tableSamples.verticalScrollBar().valueChanged.connect(self.on_table_samples_v_scroll)
        self.tableSamples.sigHover.connect(self.on_table_samples_hover)

        ## === Right ---------------------

        self.rightWidget = QtWidgets.QWidget()
        self.splitter.addWidget(self.rightWidget)
        self.rightLay = xwidgets.vlayout()
        self.rightWidget.setLayout(self.rightLay)

        ## Tests
        self.tableTests = XTableHoverView()
        self.rightLay.addWidget(self.tableTests)


        self.modelTests = OGTScheduleTestsModel(ogtDoc=self.ogtDoc)
        self.tableTests.setModel(self.modelTests)

        self.tableTests.horizontalHeader().hide()
        self.tableTests.setFixedHeight(self.TOP_HEIGHT)
        self.tableTests.sigHover.connect(self.on_table_tests_hover)


        ## Schedule
        self.tableSched = QtWidgets.QTableView()
        self.rightLay.addWidget(self.tableSched, 100)

        self.modelSchedule = OGTScheduleTableModel(ogtDoc=self.ogtDoc)
        self.tableSched.setModel(self.modelSchedule)

        self.tableSched.setMouseTracking(True)
        #self.tableSched.verticalScrollBar().hide()
        self.tableSched.verticalScrollBar().valueChanged.connect(self.on_table_sched_v_scroll)

        self.splitter.setStretchFactor(0, 2)
        self.splitter.setStretchFactor(1, 2)

    def on_butt_refresh(self):
        pass


    def on_table_samples_v_scroll(self, x):
        self.tableSched.verticalScrollBar().setValue(x)

    def on_table_sched_v_scroll(self, x):
        self.tableSamples.verticalScrollBar().setValue(x)


    def on_table_samples_hover(self, ridx, cidx):
        self.on_cell_enter(ridx, None)

    def on_table_tests_hover(self, ridx, cidx):
        self.on_cell_enter(None, cidx)

    def on_tree_samples_selection(self):
        # print "on_tree_samples_selection", "do_update"
        #self.ogtDoc.validate()
        #self.do_update()
        self.modelSchedule.hi_row = None
        idxx = self.tableSamples.selectionModel().selectedIndexes()
        if len(idxx):
            self.modelSchedule.hi_row = idxx[0].row()

        self.modelSchedule.layoutChanged.emit()



    def do_update(self):

        self.modelSamples.layoutChanged.emit()

        for c in range(0, self.modelSamples.rowCount()):
            self.tableSamples.resizeColumnToContents(c)

        sched_group = self.ogtDoc.group_from_code("LBST")
        if sched_group == None:
            return

        # first walks and get unique `tests` defined in LBST_TEST
        testsDic = {}
        for ridx, row in enumerate(sched_group.data):
            tstCell = row.get('LBST_TEST')
            if tstCell:
                tst = tstCell.value
                if not tst in testsDic:
                    testsDic[tst] = dict(test=tst, recs=[])
                testsDic[tst]["recs"].append(row)

        ## set test headings = kinda description
        self.modelTests.xheads = sorted(testsDic.keys())
        self.modelSchedule.xheads = self.modelTests.xheads

        ## create blank dict with ridx/cidx.. of None's
        self.modelSchedule.xcells = {}
        for ridx in range(0, self.modelSamples.rowCount()):
            if not ridx in self.modelSchedule.xcells:
                self.modelSchedule.xcells[ridx] = {}
                for cidx in range(0, len(self.modelSchedule.xheads)):
                    self.modelSchedule.xcells[ridx][cidx] = None




        for cidx, tst in enumerate(self.modelSchedule.xheads):
            recs = testsDic[tst]["recs"]

            for rec in recs:
                sam_id_cell = rec.get("SAMP_ID")

                ridx = self.modelSamples.row_index_from_samp_id(sam_id_cell.value)
                if ridx != None:
                    self.modelSchedule.xcells[ridx][cidx] = "YES"

        for ridx in range(0, self.modelSchedule.rowCount()):
            for cidx in range(0, self.modelSchedule.columnCount()):

                # Cell widget
                idx = self.modelSchedule.index(ridx, cidx)
                widget = self.tableSched.indexWidget(idx)
                if widget == None:
                    widget = XSchedCellWidget(ridx=ridx, cidx=cidx)
                    self.tableSched.setIndexWidget(idx, widget)
                    widget.sigEnter.connect(self.on_cell_enter)
                    widget.sigChanged.connect(self.on_check_changed)

                v = self.modelSchedule.xcells[ridx][cidx]
                if v:
                    widget.set_enabled(True)


        self.update_totals()

        self.modelTests.layoutChanged.emit()
        self.modelSchedule.layoutChanged.emit()
        self.modelSamples.layoutChanged.emit()

        self.tableTests.setRowHeight(0, self.TOP_HEIGHT)
        self.tableTests.verticalHeader().setFixedWidth(30)
        self.tableSched.verticalHeader().setFixedWidth(30)

        HEI = 24
        for ridx in range(0, self.modelSamples.rowCount()):
            self.tableSamples.setRowHeight(ridx, HEI)
            self.tableSched.setRowHeight(ridx, HEI )

    def on_check_changed(self):
        self.update_totals()
        self.modelSchedule.layoutChanged.emit()

    def update_totals(self):
        ## update totals

        self.modelSchedule.row_counts = {}
        self.modelSchedule.col_counts = {}

        for ridx in range(0, self.modelSchedule.rowCount()):
            for cidx in range(0, self.modelSchedule.columnCount()):

                # Cell widget
                widget = self.tableSched.indexWidget( self.modelSchedule.index(ridx, cidx) )
                if widget.isChecked():

                    if not ridx in self.modelSchedule.row_counts:
                        self.modelSchedule.row_counts[ridx] = 0
                    self.modelSchedule.row_counts[ridx] += 1

                    if not cidx in self.modelSchedule.col_counts:
                        self.modelSchedule.col_counts[cidx] = 0
                    self.modelSchedule.col_counts[cidx] += 1
        self.modelSchedule.layoutChanged.emit()


    def on_cell_enter(self, xridx, xcidx):
        #self.modelSchedule.set_current(ridx, cidx)

        for ridx in range(0, self.modelSchedule.rowCount()):
            for cidx in range(0, self.modelSchedule.columnCount()):
                # Cell widget
                widget = self.tableSched.indexWidget(self.modelSchedule.index(ridx, cidx))
                widget.set_hover(xridx, xcidx)

        self.modelSamples.set_current(ridx=xridx)
        self.modelSamples.layoutChanged.emit()

        self.modelTests.set_current(cidx=xcidx)
        self.modelTests.layoutChanged.emit()

class XSchedCellWidget(QtWidgets.QWidget):

    sigLeave = pyqtSignal(int, int)
    sigEnter = pyqtSignal(int, int)
    sigChanged = pyqtSignal()

    def __init__(self, parent=None, ridx=None, cidx=None):
        QtWidgets.QWidget.__init__(self)

        self.ridx = ridx
        self.cidx = cidx

        self.mainLayout = xwidgets.hlayout()
        self.setLayout(self.mainLayout)


        self.buttEna = QtWidgets.QToolButton()
        self.buttEna.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.buttEna.setCheckable(True)
        self.buttEna.setAutoRaise(True)
        self.buttEna.setFixedWidth(20)
        self.buttEna.setIcon(Ico.icon(Ico.TickOff))
        self.mainLayout.addWidget(self.buttEna, 0)
        self.buttEna.toggled.connect(self.on_enabled_toggled)
        self.buttEna.clicked.connect(self.on_butt_clicked)

        self.lblDetails = xwidgets.XLabel()
        self.mainLayout.addWidget(self.lblDetails, 2)

    def set_enabled(self, state):
        self.buttEna.setChecked(state)

    def isChecked(self):
        return self.buttEna.isChecked()

    def on_enabled_toggled(self, state=None):
        self.up_widgets()

    def on_butt_clicked(self):
        self.sigChanged.emit()

    def up_widgets(self):
        state = self.buttEna.isChecked()
        ico = Ico.TickOn if state else Ico.TickOff
        self.buttEna.setIcon(Ico.icon(ico))

        s = "background-color: #FFFA8A;" if state else ""
        self.buttEna.setStyleSheet(s)
        self.lblDetails.setStyleSheet(s)

    def enterEvent(self, ev):
        self.sigEnter.emit(self.ridx, self.cidx)

    def leaveEvent(self, ev):
        self.sigLeave.emit(self.ridx, self.cidx)

    def set_hover(self, ridx, cidx):
        #print "set_hover", ridx, self.ridx, cidx, self.cidx
        if self.buttEna.isChecked():
            return
        color = "white"

        if ridx == self.ridx or cidx == self.cidx:
            color = HOVER_COLOR

        self.lblDetails.setStyleSheet("background-color: %s" % color)
        self.buttEna.setStyleSheet("background-color: %s" % color)


class XTableHoverView(QtWidgets.QTableView):

    sigHover = pyqtSignal(int, int)

    def __init__(self, parent=None):
        QtWidgets.QTableView.__init__(self)

        self.setMouseTracking(True)

    def mouseMoveEvent(self, ev):

        idx = self.indexAt(ev.pos())
        self.sigHover.emit(idx.row(), idx.column())




# -*- coding: utf-8 -*-


from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal



from img import Ico
import xwidgets



class GroupListModel(QtCore.QAbstractItemModel):

    class C:
        data_count = 0
        group_code = 1
        group_description = 2

    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractItemModel.__init__(self)

        self.ogtDoc = ogtDoc

        self._col_labels = ["Data", "Group Code", "Descripton"]

    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()


    def columnCount(self, foo):
        return 3

    def rowCount(self, midx):
        return self.ogtDoc.groups_count

    def data(self, midx, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = midx.row()
        col = midx.column()

        if role == Qt.DisplayRole or role == Qt.EditRole:
            grp = self.ogtDoc.group_from_index(row)

            if col == self.C.group_code:
                return grp.group_code

            if col == self.C.group_description:
                return grp.group_description

            if col == self.C.data_count:
                return grp.data_rows_count

            return "?"


        if role == Qt.DecorationRole:
            if col == self.C.group_code:
                return Ico.icon(Ico.Group)

        if role == Qt.FontRole:
            if col == self.C.group_code:
                f = QtGui.QFont()
                f.setBold(True)
                f.setFamily("monospace")
                return f

        if role == Qt.TextAlignmentRole:
            return Qt.AlignRight if col == 0 else Qt.AlignLeft

        if False and role == Qt.BackgroundColorRole:
            cell = self.ogtDoc.group_from_index(row)[col]
            bg = cell.get_bg()
            if len(self.ogtGroup.data_cell(row, col).errors) > 0:
                pass
            return QtWidgets.QColor(bg)


        return QtCore.QVariant()


    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            return QtCore.QVariant(self._col_labels[idx])

        if role == Qt.TextAlignmentRole and orient == Qt.Horizontal:
            return Qt.AlignRight if idx == 0 else Qt.AlignLeft

        return QtCore.QVariant()


class GroupsListWidget( QtWidgets.QWidget ):

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False

        self.setMinimumWidth(300)

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)


        self.tree = QtWidgets.QTreeView()
        self.mainLayout.addWidget(self.tree)

        self.model = GroupListModel(ogtDoc=ogtDoc)
        self.tree.setModel(self.model)

        self.tree.setMinimumWidth(300)
        self.tree.setRootIsDecorated(False)
        self.tree.setExpandsOnDoubleClick(False)
        self.tree.header().setStretchLastSection(True)

        self.tree.setColumnWidth(GroupListModel.C.data_count, 40)


    def do_update(self):
        self.model.layoutChanged.emit()



    def DEADload_projects(self, sub_dir=None):
        return
        files_list, err = ags4.examples_list()
        if err:
            pass #TODO
        self.tree.clear()

        for fd in files_list:
            file_name = fd["file_name"]
            item = QtWidgets.QTreeWidgetItem()
            item.setText(C_EG.file_name, file_name)
            item.setIcon(C_EG.file_name, Ico.icon(Ico.Ags4))
            f = item.font(C_EG.file_name)
            f.setBold(True)
            item.setFont(C_EG.file_name, f)
            self.tree.addTopLevelItem(item)



    def deadon_tree_item_clicked(self, item, col):

        file_name = str(item.text(C_EG.file_name))
        self.sigFileSelected.emit(file_name)
